# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "marimo>=0.19.0",
#     "pyzmq",
# ]
# ///

# NOTE: Use the command palette (cmd+shift+p) and search "marimo start in editor" to generate the live preview.

import marimo

__generated_with = "0.19.11"
app = marimo.App(width="full")

with app.setup:
    import numpy as np
    from mcp_hmm.hmm_xiang2023 import (
        generate_windows,
        generate_outcome,
        WindowableHMM,
        DebuggingMonitor,
        plot_results,
        plot_hmm_matrices,
        visualize_agent_contributions_sampdist,
        AgentPair,
        TOKENS,
        TOKEN_COLORS,
        make_null_actions_generator,
        generate_windows_with_actions_generator
    )
    import networkx as nx
    import polars as pl
    from polars import lit, col
    import marimo as mo

    from sklearn.model_selection import KFold, LeaveOneGroupOut

    import seaborn as sns
    from matplotlib import pyplot as plt
    from scipy.optimize import minimize


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Cross-validation
    """)
    return


@app.cell(hide_code=True)
def _():
    # discriminability = mo.ui.slider(start=0, stop=1, value=.5, step=.1, label="Discriminability")
    discriminability_noise = mo.ui.slider(start=0, stop=1, value=.2, step=.1, label="Discriminability Noise")
    A_competence = mo.ui.slider(start=0, stop=10, value=4, step=.5, label="A Competence")
    B_competence = mo.ui.slider(start=0, stop=10, value=6, step=.5, label="B Competence")

    mo.vstack([ discriminability_noise, A_competence, B_competence ])
    return A_competence, B_competence, discriminability_noise


@app.cell(hide_code=True)
def _(A_competence, B_competence, discriminability_noise):
    np.random.seed(4)
    pair = AgentPair(discriminability_noise=discriminability_noise.value, A_competence=A_competence.value, B_competence=B_competence.value)

    visualize_agent_contributions_sampdist(pair)
    return (pair,)


@app.cell(hide_code=True)
def _(accuracy_results):
    show_accuracy(accuracy_results)
    return


@app.function(hide_code=True)
def get_window_types(windows):
    _, unique_indices = np.unique(np.apply_along_axis(str, 1, windows), return_inverse=True)
    return unique_indices


@app.cell(hide_code=True)
def _(pair):
    n_windows = 100

    np.random.seed(4)
    fwd_windows = np.array(generate_windows(pair, num_windows=n_windows))
    bwd_windows = np.flip(fwd_windows)  # for the backwards model, we flip each window
    _bwd_windows_types = get_window_types(bwd_windows)

    test_scores = []
    train_scores = []
    test_scores_b = []
    train_scores_b = []
    for i, (train, test) in enumerate(LeaveOneGroupOut().split(bwd_windows, groups=_bwd_windows_types)):
        np.random.seed(i)

        train_windows = bwd_windows[train]
        test_windows = bwd_windows[test]

        bwd_model = WindowableHMM(n_states=10, max_iter=120)
        for win in train_windows:
            bwd_model.add_to_history(win)
        bwd_model.fit()

        print(f"{bwd_model.model.monitor_.iter} iterations out of {bwd_model.model.monitor_.n_iter} max")

        train_scores.append(bwd_model.accuracy(train_windows, backward_learning=True).reshape(-1))
        test_scores.append(bwd_model.accuracy(test_windows, backward_learning=True).reshape(-1))

        train_scores_b.append(bwd_model.accuracy_about_stronger_agent(train_windows, backward_learning=True).reshape(-1))
        test_scores_b.append(bwd_model.accuracy_about_stronger_agent(test_windows, backward_learning=True).reshape(-1))
    return (
        bwd_windows,
        n_windows,
        test_scores,
        test_scores_b,
        train_scores,
        train_scores_b,
    )


@app.function
def tidy_scores(folds):
    fold_idxs = []
    window_idxs = []
    scores = []
    for fold_idx, fold in enumerate(folds):
        for window_idx, score in enumerate(fold):
            fold_idxs.append(fold_idx)
            window_idxs.append(window_idx)
            scores.append(score)
    return pl.DataFrame({
        'fold': fold_idxs,
        'window_idx': window_idxs,
        'score': scores
    })


@app.cell(hide_code=True)
def _(test_scores, test_scores_b, train_scores, train_scores_b):
    accuracy_results = pl.concat([
        tidy_scores(train_scores).with_columns(label=lit('train'), target=lit('both agents')),
        tidy_scores(test_scores).with_columns(label=lit('test'), target=lit('both agents')),
        tidy_scores(train_scores_b).with_columns(label=lit('train'), target=lit('stronger agent')),
        tidy_scores(test_scores_b).with_columns(label=lit('test'), target=lit('stronger agent'))
    ])
    return (accuracy_results,)


@app.function(hide_code=True)
def show_accuracy(acc):
    return mo.vstack([
        sns.catplot(
            acc.to_pandas(),
            kind='bar',
            x='fold',
            hue='label',
            y='score',
            col='target'
        ).map_dataframe(
            sns.stripplot,
            x='fold',
            hue='label',
            y='score',
            dodge='auto',
            alpha=.1
        ).set(ylim=[0, 1]).map(plt.axhline, y=.5, ls='--', c='gray'),
        acc.group_by('label', 'target', maintain_order=True)
            .agg(mean_score = col('score').drop_nans().mean())
            .pivot(values='mean_score', on='target', maintain_order=True)
    ])


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Inspecting trajectory of hidden weights during fit
    """)
    return


@app.cell(hide_code=True)
def _(bwd_windows):
    np.random.seed(2)
    _n_states = 10
    # TI = trajectory inspector
    TI_bwd_model = WindowableHMM(n_states=_n_states, max_iter=120, save_props_while_fitting=['startprob_', 'transmat_', 'emissionprob_'], custom_model_args={
        # randomize the transmat priors to try to help more varied solutions arise (randomizing the startprob_prior doesn't seem to have an impacts, and can only initialize all the emission probs to the same value)
        'transmat_prior': np.random.rand(_n_states, _n_states)
    })
    for _win in bwd_windows:
        TI_bwd_model.add_to_history(_win)
    TI_bwd_model.fit()
    return (TI_bwd_model,)


@app.function
def loss_plot(model, frame=0):
    return sns.relplot(data=pl.DataFrame({
        'Iteration': np.arange(1, len(model.convergence_history)),
        '-logP': -model.convergence_history[1:]
    }), x='Iteration', y='-logP', kind='line', height=3).map(plt.axvline, x=frame)


@app.function
def make_heatmap_trajectory_inspector(model):
    _order = order_hiddens_by_emission_probs(model, model.fitting_history[-1])
    # ^ define order based on the final iteration so that we see the trajectory towards that

    matrix_figs = [
        plot_hmm_matrices(
            model.model.startprob_prior_.reshape(model.n_states, 1),
            model.model.transmat_prior_,
            model.model.emissionprob_prior_,
            order=_order
        )
    ] + [
        plot_hmm_matrices(
            _record['startprob_'].reshape(model.n_states, 1),
            _record['transmat_'],
            _record['emissionprob_'],
            order=_order
        )
        for _record in model.fitting_history
    ]

    def show(frame_slider):
        return mo.vstack([
            mo.hstack([ frame_slider, mo.md(f"Frame {frame_slider.value}/{frame_slider.stop} (0 = prior)") ]),
            mo.hstack([ matrix_figs[frame_slider.value], loss_plot(model, frame_slider.value) ])
        ])

    frame_slider = mo.ui.slider(start=0, stop=len(matrix_figs) - 1, step=1, value = len(matrix_figs) - 1, label="Frame")

    return show, frame_slider


@app.cell
def _(TI_bwd_model):
    hmTI, hmTI_frame_slider = make_heatmap_trajectory_inspector(TI_bwd_model)
    return hmTI, hmTI_frame_slider


@app.cell
def _(hmTI, hmTI_frame_slider):
    hmTI(hmTI_frame_slider)
    return


@app.function(hide_code=True)
def order_hiddens_by_emission_probs(model: WindowableHMM, reference_record: dict):
    """
    Example usage with plot_hmm_matrices:
    _record = TI_bwd_model.fitting_history[-1]; plot_hmm_matrices(_record['startprob_'].reshape(TI_bwd_model.n_states, 1), _record['transmat_'], _record['emissionprob_'], order=order_hiddens_by_emission_probs(TI_bwd_model, _record))
    """
    from functools import cmp_to_key

    def compare_by_emission_probs(hidden_a, hidden_b):
        emission_idxs = np.arange(0, model.model.n_features)
        hidden_a_emissions = reference_record['emissionprob_'][hidden_a]
        hidden_b_emissions = reference_record['emissionprob_'][hidden_b]
        correlation_a = np.correlate(hidden_a_emissions, emission_idxs)
        correlation_b = np.correlate(hidden_b_emissions, emission_idxs)

        # if np.std(hidden_a_emissions) < .001 or np.std(hidden_b_emissions) < .001:
        #     return 1
        # ^ doesn't always work as intended (was meant to push the states not used for emissions to the bottom of the matrix)
        if correlation_a > correlation_b:
            return 1
        elif correlation_a < correlation_b:
            return -1
        else:
            return 0

    order = sorted(
        np.arange(0, model.n_states),
        key=cmp_to_key(compare_by_emission_probs)
    )
    return order


@app.function(hide_code=True)
def visualize_markov_chain(startprob, transmat, emissionprob, model: WindowableHMM, threshold=0.05, order=None):
    """Visualize a fit WindowableHMM's hidden states as a Markov chain directed graph.

    Args:
        model: A fit WindowableHMM instance (to extract n_states).
        startprob, transmat, emissionprob: Matrices (cf. plot_hmm_matrices)
        threshold: Minimum transition probability to display an edge.
        order: List of hidden state indices to include (and their display order).
               If None, all states are shown.
    """

    if order is None:
        states = list(range(model.n_states))
    else:
        states = list(order)

    # Compute dominant emission token per state
    dominant_token = {}
    for s in states:
        dominant_token[s] = TOKENS[np.argmax(emissionprob[s])]

    # Build directed graph
    G = nx.DiGraph()
    for s in states:
        G.add_node(s)

    self_loop_edges = {}
    normal_edges = {}
    for i in states:
        for j in states:
            prob = transmat[i, j]
            if prob >= threshold:
                if i == j:
                    self_loop_edges[(i, j)] = prob
                else:
                    normal_edges[(i, j)] = prob
                G.add_edge(i, j)

    fig, ax = plt.subplots(figsize=(14, 10))
    pos = nx.spring_layout(G, k=2, iterations=50, seed=42)

    # Node sizes scaled by start probability
    min_size, max_size = 2000, 5000
    sp = np.array([startprob[s] for s in G.nodes()])
    if sp.max() > sp.min():
        node_sizes = min_size + (max_size - min_size) * (sp - sp.min()) / (sp.max() - sp.min())
    else:
        node_sizes = np.full(len(sp), (min_size + max_size) / 2)

    node_colors = [TOKEN_COLORS.get(dominant_token[s], "#bdc3c7") for s in G.nodes()]

    nx.draw_networkx_nodes(G, pos, node_color=node_colors,
                           node_size=node_sizes, alpha=0.9, ax=ax)

    labels = {s: f"S{s}\n{dominant_token[s]}" for s in G.nodes()}
    nx.draw_networkx_labels(G, pos, labels=labels, font_size=10,
                            font_weight='bold', font_color='white', ax=ax)

    # Draw normal edges
    if normal_edges:
        nx.draw_networkx_edges(
            G, pos,
            edgelist=list(normal_edges.keys()),
            width=[5 * normal_edges[e] for e in normal_edges],
            alpha=0.7,
            edge_color='#555555',
            connectionstyle='arc3,rad=0.2',
            arrows=True,
            arrowsize=20,
            arrowstyle='-|>',
            node_size=max_size,
            ax=ax,
            min_source_margin=15,
            min_target_margin=15
        )

    # Draw self-loops with higher curvature
    if self_loop_edges:
        nx.draw_networkx_edges(
            G, pos,
            edgelist=list(self_loop_edges.keys()),
            width=[5 * self_loop_edges[e] for e in self_loop_edges],
            alpha=0.7,
            edge_color='#555555',
            arrows=True,
            arrowsize=20,
            arrowstyle='-|>',
            node_size=max_size,
            ax=ax,
        )

    # Edge labels
    all_edges = {**normal_edges, **self_loop_edges}
    edge_labels = {e: f"{p:.2f}" for e, p in all_edges.items()}
    label_bbox = dict(boxstyle='round,pad=0.2', facecolor='white',
                      alpha=0.8, edgecolor='gray', linewidth=0.5)
    nx.draw_networkx_edge_labels(
        G, pos,
        edge_labels=edge_labels,
        font_size=8,
        bbox=label_bbox,
        ax=ax,
        connectionstyle='arc3,rad=0.2'
    )

    ax.set_title(f"Hidden State Markov Chain (threshold={threshold:.2f})",
                 fontsize=14, fontweight='bold', pad=20)
    ax.axis('off')
    plt.tight_layout()

    return fig


@app.function
def make_markov_chain_trajectory_inspector(model):
    _order = order_hiddens_by_emission_probs(
        model, model.fitting_history[-1]
    )

    def make_figs(threshold):
        chain_figs = [
            visualize_markov_chain(
                model.model.startprob_prior_.reshape(model.n_states, 1),
                model.model.transmat_prior_,
                model.model.emissionprob_prior_,
                model,
                threshold=threshold,
                order=_order,
            )
        ] + [
            visualize_markov_chain(
                _record["startprob_"].reshape(model.n_states, 1),
                _record["transmat_"],
                _record["emissionprob_"],
                model,
                threshold=threshold,
                order=_order,
            )
            for _record in model.fitting_history
        ]
        return chain_figs

        # order dictates the drawing order, which seems to make it match the heatmap order better
        # Note that should only follow the path for two hops, since the model's predictions after two steps are not relevant (not trained, and it's not defined how they'd be useful i.e. undefined behavior)
        # And note that nodes are labeled with their **dominant** (i.e., most likely) emission, even though they may have multiple possible emissions (and that ambiguity could potentially be informative).

    def show(frame_slider, threshold_slider):
        chain_figs = make_figs(threshold_slider.value)
        return mo.vstack([
            threshold_slider,
            mo.hstack([ frame_slider, mo.md(f"Frame {frame_slider.value}/{frame_slider.stop} (0 = prior)") ]),
            mo.hstack([ chain_figs[frame_slider.value], loss_plot(model, frame_slider.value) ])
        ])

    frame_slider = mo.ui.slider(start=0, stop=len(model.fitting_history) - 1 + 1, step=1, label="Frame", value = len(model.fitting_history) - 1 + 1)

    threshold_slider = mo.ui.slider(start=0.01, stop=0.5, step=0.01, value=0.15, label="Edge threshold")

    return show, frame_slider, threshold_slider


@app.cell
def _(TI_bwd_model):
    mcTI, _, mcTI_threshold_slider = make_markov_chain_trajectory_inspector(TI_bwd_model)
    return mcTI, mcTI_threshold_slider


@app.cell
def _(hmTI_frame_slider, mcTI, mcTI_threshold_slider):
    mcTI(hmTI_frame_slider, mcTI_threshold_slider)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Inferring competence from learned transition probabilities

    Our model has that:
    - P(lift | X_try, Y_try) ~ (Competence(X) + Competence(Y)) >= 5
    - P(lift | X_try, Y_wait) ~ Competence(X) >= 5
    - P(lift | X_wait, Y_wait) ~ 0
    - P(fail | ...) ~ 1 - P(lift | ...)

    with X and Y being A and B in any order,  with the conditioning events in no particular order, and ... being any of the three conditions. Importantly, `discriminability_noise` doesn't change these relationships, it just introduces noise.

    So then let Competence(X) be inferred by Bayesian inference on the probabilities learned by the HMM (where X can be A or B). If the HMM is backwards-learning (i.e., outcome comes first), I think we can just apply Bayes' rule to compute P(lift | ...) and P(fail | ...).
    """)
    return


@app.function
# Step 1: Extract P(lift | actions) from the backward HMM via Bayes' rule
# The backward model sees [outcome, action1, action2].
#
# To compute joint P(O1=a1, O2=a2 | O0=outcome), we must propagate through
# hidden states properly (cf. _joint_prob in hmm_xiang2023.py), NOT multiply
# marginal predict() outputs which would ignore hidden-state dependence.

def infer_lift_prob(bwd_model, pair):
    _model = bwd_model.model
    _tok2idx = bwd_model.token_to_idx

    # P(outcome) from the HMM's learned start/emission matrices
    # (first token in backward sequences is the outcome)
    _p_outcome = {}
    for _outcome in ["lift", "fail"]:
        _oidx = _tok2idx[_outcome]
        _p_outcome[_outcome] = (_model.startprob_ * _model.emissionprob_[:, _oidx]).sum()

    # Compute joint P(O1=a1, O2=a2 | O0=outcome) properly through hidden states:
    #   1. Get state dist after observing outcome: score_samples([outcome])
    #   2. Transition to t=1 states, weight by P(emit a1)
    #   3. Transition to t=2 states, weight by P(emit a2)
    #   (This is the _joint_prob pattern from hmm_xiang2023.py)
    def _joint_prob_given_outcome(outcome, a1, a2):
        _obs = bwd_model._tokens_to_obs([outcome])
        _, _posteriors = _model.score_samples(_obs, [len(_obs)])
        _state_dist_0 = _posteriors[-1]

        _a1_idx = _tok2idx[a1]
        _a2_idx = _tok2idx[a2]

        _state_dist_1 = _state_dist_0 @ _model.transmat_
        _weighted = _state_dist_1 * _model.emissionprob_[:, _a1_idx]
        _state_dist_2 = _weighted @ _model.transmat_
        return (_state_dist_2 * _model.emissionprob_[:, _a2_idx]).sum()

    # Conditions: unordered action pairs
    _conditions = [
        ("A_try", "B_try"),
        ("A_try", "B_wait"),
        ("B_try", "A_wait"),
        ("A_wait", "B_wait"),
    ]

    _hmm_p_lift = {}
    for _a1, _a2 in _conditions:
        _label = f"{_a1}, {_a2}"
        _p_actions_given = {}
        for _outcome in ["lift", "fail"]:
            # ordering 1: a1 then a2
            _p_order1 = _joint_prob_given_outcome(_outcome, _a1, _a2)

            if _a1 != _a2:  # (defensive programming)
                # ordering 2: a2 then a1
                _p_order2 = _joint_prob_given_outcome(_outcome, _a2, _a1)
                _p_actions_given[_outcome] = _p_order1 + _p_order2
            else:
                _p_actions_given[_outcome] = _p_order1

        # Bayes' rule: P(lift | {a1, a2}) = P({a1, a2} | lift) * P(lift) / P({a1, a2})
        _p_actions = sum(_p_actions_given[o] * _p_outcome[o] for o in ["lift", "fail"])
        if _p_actions > 0:
            _hmm_p_lift[_label] = _p_actions_given["lift"] * _p_outcome["lift"] / _p_actions
        else:
            _hmm_p_lift[_label] = float("nan")

    # Ground truth via Monte Carlo
    _n_mc = 10000
    _gt_p_lift = {}
    for _a1, _a2 in _conditions:
        _label = f"{_a1}, {_a2}"
        _lifts = 0
        for _ in range(_n_mc):
            _lifts += 1 if generate_outcome(pair, _a1, _a2) == "lift" else 0
        _gt_p_lift[_label] = _lifts / _n_mc

    lift_prob = {
        "conditions": [f"{a1}, {a2}" for a1, a2 in _conditions],
        "hmm_p_lift": _hmm_p_lift,
        "gt_p_lift": _gt_p_lift,
        "p_outcome": _p_outcome,
        "noise": 0.5,
        # ^ don't just feed in discriminability_noise.value -- it should be inferred by the participant model from the data, OR a free parameter set to optimize fit to real participant data. it's cheating if the model of the participant knows generation parameters! should only have access to observations
    }
    return lift_prob


@app.cell
def _(TI_bwd_model, pair):
    lift_prob = infer_lift_prob(TI_bwd_model, pair)
    return (lift_prob,)


@app.cell
def _(lift_prob, pair):
    # Step 2: Comparison table + competence inversion
    _lp = lift_prob

    _comparison_df = pl.DataFrame({
        "Condition": _lp["conditions"],
        "HMM P(lift)": [_lp["hmm_p_lift"][c] for c in _lp["conditions"]],
        "Ground Truth P(lift)": [_lp["gt_p_lift"][c] for c in _lp["conditions"]],
    }).with_columns(
        Difference=(col("HMM P(lift)") - col("Ground Truth P(lift)"))
    )

    # Invert solo-try conditions to estimate competence
    # Effort ~ Beta(alpha, 1), so p_lift = P(C*E >= 5) = 1 - (5/C)^alpha when C > 5
    # Inverting: C = 5 / (1 - p_lift)^(1/alpha)
    _noise = _lp["noise"]
    _alpha = 1.0 / max(_noise**2, 0.00001)

    def _invert_competence(p_lift):
        if p_lift == 0:
            return None  # can only say C <= 5
        if p_lift == 1:
            return float("inf")  # always succeeds solo, C unbounded above
        return 5.0 / (1.0 - p_lift) ** (1.0 / _alpha)

    _agents = ["A", "B"]
    _true_comp = [pair.A_competence, pair.B_competence]
    _solo_conditions = ["A_try, B_wait", "B_try, A_wait"]
    _hmm_inferred = [_invert_competence(_lp["hmm_p_lift"][c]) for c in _solo_conditions]
    _gt_inferred = [_invert_competence(_lp["gt_p_lift"][c]) for c in _solo_conditions]

    _competence_rows = []
    for _i, _agent in enumerate(_agents):
        _competence_rows.append({
            "Agent": _agent,
            "True Competence": _true_comp[_i],
            "GT-Inferred": _gt_inferred[_i] if _gt_inferred[_i] is not None else float("nan"),
            "HMM-Inferred": _hmm_inferred[_i] if _hmm_inferred[_i] is not None else float("nan"),
            "Solo P(lift) [HMM]": _lp["hmm_p_lift"][_solo_conditions[_i]],
            "Solo P(lift) [GT]": _lp["gt_p_lift"][_solo_conditions[_i]],
        })
    competence_results = pl.DataFrame(_competence_rows)

    mo.vstack([
        mo.md(f"""
    ### P(lift | actions): HMM vs Ground Truth

    HMM marginal: P(lift) = {_lp['p_outcome']['lift']:.3f}, P(fail) = {_lp['p_outcome']['fail']:.3f}

    Effort distribution: Beta(alpha={_alpha:.1f}, 1), noise={_noise}
        """),
        _comparison_df,
        mo.md("### Competence inversion from solo-try conditions"),
        mo.md(f"Formula: C = 5 / (1 - P(lift))^(1/{_alpha:.1f}). NaN means P(lift) = 0, so C <= 5."),
        competence_results,
    ])
    return


@app.function
# Step 3: Joint competence inference using solo AND joint-try conditions
# Solo: P(lift | X_try, Y_wait) = 1 - (5/C_X)^alpha  (when C_X > 5, else 0)
# Joint: P(lift | A_try, B_try) = P(C_A*E_A + C_B*E_B >= 5)  (no closed form)
#
# We find (C_A, C_B) that best fits all three observed lift probabilities.

def infer_competence_from_lift_prob(_lp, pair):
    _noise = _lp["noise"]
    _alpha = 1.0 / max(_noise**2, 0.00001)
    _n_mc = 10000

    def _predict_p_lift_solo(c, alpha):
        if c <= 5:
            return 0.0
        return 1.0 - (5.0 / c) ** alpha

    def _predict_p_lift_joint(c_a, c_b, alpha, n_mc):
        _e_a = np.random.beta(alpha, 1, size=n_mc)
        _e_b = np.random.beta(alpha, 1, size=n_mc)
        return np.mean(c_a * _e_a + c_b * _e_b >= 5)

    _p_lift_a_solo = _lp["hmm_p_lift"]["A_try, B_wait"]
    _p_lift_b_solo = _lp["hmm_p_lift"]["B_try, A_wait"]
    _p_lift_joint = _lp["hmm_p_lift"]["A_try, B_try"]

    def _loss(params):
        _c_a, _c_b = params
        if _c_a < 0 or _c_b < 0:
            return 1e6
        _pred_a = _predict_p_lift_solo(_c_a, _alpha)
        _pred_b = _predict_p_lift_solo(_c_b, _alpha)
        np.random.seed(0)
        _pred_joint = _predict_p_lift_joint(_c_a, _c_b, _alpha, _n_mc)
        return (
            (_pred_a - _p_lift_a_solo) ** 2
            + (_pred_b - _p_lift_b_solo) ** 2
            + (_pred_joint - _p_lift_joint) ** 2
        )

    # Try multiple starting points to avoid local minima
    _best = None
    for _c_a_init in [2, 4, 6, 8]:
        for _c_b_init in [2, 4, 6, 8]:
            _result = minimize(_loss, [_c_a_init, _c_b_init], method="Nelder-Mead")
            if _best is None or _result.fun < _best.fun:
                _best = _result

    _c_a_best, _c_b_best = _best.x

    joint_competence_results = pl.DataFrame({
        "Agent": ["A", "B"],
        "True Competence": [pair.A_competence, pair.B_competence],
        "Solo+Joint-Inferred": [_c_a_best, _c_b_best],  # i.e., inferred from both solo and joint attempts
    })

    original_predicted_p = {
        'A_solo': _p_lift_a_solo,
        'B_solo': _p_lift_b_solo,
        'joint': _p_lift_joint
    }

    optimized_predicted_p = {
        'A_solo': _predict_p_lift_solo(_c_a_best, _alpha),
        'B_solo': _predict_p_lift_solo(_c_b_best, _alpha),
        'joint': _predict_p_lift_joint(_c_a_best, _c_b_best, _alpha, _n_mc)
    }
    return joint_competence_results, original_predicted_p, optimized_predicted_p


@app.cell
def _(lift_prob, pair):
    np.random.seed(0)
    joint_competence_results, original_predicted_p, optimized_predicted_p = infer_competence_from_lift_prob(lift_prob, pair)

    # Show predicted vs observed lift probs for the joint fit
    mo.vstack([
        mo.md(f"""
    ### Joint competence inference (using solo + both-try conditions)

    Jointly optimizes (C_A, C_B) to fit three HMM-derived lift probabilities:
    - P(lift | A_try, B_wait) = {original_predicted_p['A_solo']:.3f} (predicted: {optimized_predicted_p['A_solo']:.3f})
    - P(lift | B_try, A_wait) = {original_predicted_p['B_solo']:.3f} (predicted: {optimized_predicted_p['B_solo']:.3f})
    - P(lift | A_try, B_try) = {original_predicted_p['joint']:.3f} (predicted: {optimized_predicted_p['joint']:.3f})
        """),
        joint_competence_results,
    ])
    return (joint_competence_results,)


@app.function
# Step 4: Bar chart comparing true vs inferred competence (solo and solo+joint)

def visualize_competence_results(joint_competence_results, pair):
    _plot_df = joint_competence_results.unpivot(
        on=["True Competence", "Solo+Joint-Inferred"],
        index="Agent",
        variable_name="Source",
        value_name="Competence",
    )

    _fig, _ax = plt.subplots(figsize=(7, 4))
    _x = np.arange(2)
    _width = 0.25
    _sources = ["True Competence", "Solo+Joint-Inferred"]
    _colors = ["#2ecc71", "#e67e22"]

    for _i, (_source, _color) in enumerate(zip(_sources, _colors)):
        _vals = _plot_df.filter(col("Source") == _source)["Competence"].to_numpy()
        _ax.bar(_x + _i * _width, _vals, _width, label=_source, color=_color)

    _ax.axhline(y=5, ls="--", color="gray", label="Lift threshold")
    _ax.set_xticks(_x + _width)
    _ax.set_xticklabels(["A", "B"])
    _ax.set_ylabel("Competence")
    _ax.set_title(f"Competence Inference (A={pair.A_competence}, B={pair.B_competence})")
    _ax.legend()
    _fig.tight_layout()
    return _fig


@app.cell
def _(joint_competence_results, pair):
    visualize_competence_results(joint_competence_results, pair)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Cross-prediction from null to intact data

    Null data so defined that this tests that the model needs to learn the temporal contingencies of the actions.
    """)
    return


@app.cell
def _(bwd_windows, n_windows, pair):
    null_actions_gen, _, _ = make_null_actions_generator()

    np.random.seed(4)
    null_fwd_windows = np.array(generate_windows_with_actions_generator(pair, n_windows, null_actions_gen))
    null_bwd_windows = np.flip(null_fwd_windows)  # for the backwards model, we flip each window
    _null_bwd_windows_types = get_window_types(null_bwd_windows)

    _test_scores = []
    _train_scores = []
    _test_scores_b = []
    _train_scores_b = []
    for _i, (_train, _test) in enumerate(LeaveOneGroupOut().split(null_bwd_windows, groups=_null_bwd_windows_types)):
        np.random.seed(_i)

        _train_windows = null_bwd_windows[_train]

        _rng = np.random.default_rng()
        _test_scores_iteration_means = []
        _test_scores_b_iteration_means = []
        for _j in range(10):
            print(f"Fold {_i}, Iteration {_j}")
            _test_windows = _rng.choice(bwd_windows, size=len(_test), replace=False)
            # ^ could use ALL intact data, but for the sake of matching the precision of the scores to the regular version, sample the same number as in the selected test set.

            _bwd_model = WindowableHMM(n_states=10, max_iter=120)
            for _win in _train_windows:
                _bwd_model.add_to_history(_win)
            _bwd_model.fit()

            print(f"{_bwd_model.model.monitor_.iter} iterations out of {_bwd_model.model.monitor_.n_iter} max")

            _test_scores_iteration_means.append(_bwd_model.accuracy(_test_windows, backward_learning=True).reshape(-1).mean())
            _test_scores_b_iteration_means.append(_bwd_model.accuracy_about_stronger_agent(_test_windows, backward_learning=True).reshape(-1).mean())

        _train_scores.append(_bwd_model.accuracy(_train_windows, backward_learning=True).reshape(-1))
        _test_scores.append(_test_scores_iteration_means)

        _train_scores_b.append(_bwd_model.accuracy_about_stronger_agent(_train_windows, backward_learning=True).reshape(-1))
        _test_scores_b.append(_test_scores_b_iteration_means)

    null_cp_accuracy_results = pl.concat([
        tidy_scores(_train_scores).with_columns(label=lit('train'), target=lit('both agents')),
        tidy_scores(_test_scores).with_columns(label=lit('test'), target=lit('both agents')),
        tidy_scores(_train_scores_b).with_columns(label=lit('train'), target=lit('stronger agent')),
        tidy_scores(_test_scores_b).with_columns(label=lit('test'), target=lit('stronger agent'))
    ])
    return (null_cp_accuracy_results,)


@app.cell
def _(null_cp_accuracy_results):
    show_accuracy(null_cp_accuracy_results)
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "marimo>=0.19.0",
#     "pyzmq",
# ]
# ///

# NOTE: Use the command palette (cmd+shift+p) and search "marimo start in editor" to generate the live preview.

import marimo

__generated_with = "0.19.11"
app = marimo.App(width="full")

with app.setup(hide_code=True):
    import marimo as mo
    import numpy as np
    import matplotlib.pyplot as plt
    import seaborn as sns
    import polars as pl
    import matplotlib.patches as mpatches
    from collections import Counter, defaultdict
    import warnings

    warnings.filterwarnings("ignore")

    import hmmlearn.vhmm
    from hmmlearn.base import ConvergenceMonitor

    SEED = 4

    TOKENS = ["A_try", "A_wait", "B_try", "B_wait", "lift", "fail"]

    TOKEN_COLORS = {
        "A_try": "#e67e22",
        "B_try": "#e67e22",
        "A_wait": "#95a5a6",
        "B_wait": "#95a5a6",
        "lift": "#2ecc71",
        "fail": "#e74c3c",
    }


@app.class_definition(hide_code=True)
class AgentPair:
    # A = A_competence * A_effort
    # A_effort ~ U(0,1) or = 1
    # B = B_competence * B_effort
    # B_effort ~ U(0,1) or = 1
    # for A_try,B_try,lift / B_try,A_try,lift:
    # easiest credit assignment is when Pr[(A+B)>5]>0 but |Pr[A>5]-Pr[B>5]|=~1 (e.g. Pr[A>5]=0 and Pr[B>5] is large. here, B is the one that is responsible)
    # impossible (i.e., not able to be better than chance) credit assignment is when Pr[(A+B)>5]>0 but |Pr[A>5]-Pr[B>5]|=~0

    def __init__(self, A_competence=None, B_competence=None, discriminability=0.5, discriminability_noise=0.5):
        """
        discriminability: 0 = impossible credit assignment, 1 = easiest
        discriminability_noise: 0 = no variability in effort, 1 = maximum variability in effort
        """

        min_lift_competence = 5

        if A_competence is None:
            self.A_competence = min_lift_competence * (1 - discriminability)
        else:
            self.A_competence = A_competence

        if B_competence is None:
            self.B_competence = min_lift_competence * (1 + discriminability)
        else:
            self.B_competence = B_competence

        self.discriminability = discriminability
        self.discriminability_noise = discriminability_noise

    def _generate_effort(self):
        # hand-crafted in desmos such that discriminability_noise=1 -> effort ~ U(0,1), while discriminability_noise=0 -> effort = 1, and it interpolates between those
        beta = 1.0
        alpha = 1.0 / np.max([self.discriminability_noise**2.0, 0.00001])
        return np.random.beta(alpha, beta)

    def A_effort(self):
        return self._generate_effort()

    def B_effort(self):
        return self._generate_effort()


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Use (optionally) disjoint-windowed HMM to model box-lifting action sequences.

    Notes:
    - 2/5/26: it seems that maybe probability of outcomes are being underrated when they are unlikely in general, even when they are likely in context. i.e. **HMM seems sensitive to base rates and not just conditional probability/exhibits the base rate fallacy!** -- ~~IDEA don't have the input sequences be random; instead, just show all combinations the same amount of times~~; IDEA to diagnose if base rate fallacy is driving the weird results (see sanity checks output), calculate the frequency of each n-gram which the sanity checks are trying to model (as it varies slightly per each random seed).
    - 2/8/26 (currently, A=0 and B=10, and SEED=3): **backward model provides evidence of being able to de-alias outcomes (i.e., assign credit to/infer which agent tried)**! however, I'm also finding it's learning some weird statistical patterns (that I believe aren't there):
        - e.g., the backwards model is completely sure that in `?,B_wait,fail`, `?` is `A_try`, even though it should be equally likely as `A_wait`. maybe this is due to a base rate fallacy? (since WW is half as likely as TW).
        - similarly, for `?,fail` it should only infer `Pr[B_try] = 0`, and `Pr[A_wait]` should match `Pr[A_try]`, and `Pr[B_wait]` should be relatively high (not sure if should be higher than `Pr[A_try]`, though; would need to do the math)
        - there seems to be a relationship between A_try and fail, even though A_try/A_wait does not affect outcome; maybe it's due to some weird dependency like B_try-(+)>lift, A_try-(-)>B_try, thus A_try-(-)>lift; **there should be no relationship between A's and B's actions**. ~~***maybe I should feed in both agents' actions at once (i.e. encode the 4 unordered combinations A_try/wait,B_try/wait as four tokens)?**~~ no, that would actually just be making the problem for the HMM trivial. its job is to infer across time
    - 2/8/26: ~~should I be keeping the sequences as separate windows as I am currently doing? maybe I mean to/should have all the observations come in as a continuous sequence, though I'm not sure if this should make a meaningful difference, and if anything, feeding sequences as separate windows should make things easier for the HMM. _though maybe this is actually leading to the weird statistical patterns somehow (by not allowing the HMM to learn on its own what contingencies are not relevant)?_~~
        - 2/8/26: update, yeah I'll combine the sequences both to allow the HMM learn on its own what contingencies are irrelevant, and to make interpreting the predictions of illegal actions make some more sense (i.e., as I'm loosening the structure of the input data). the weird statistical patterns it's learning are still there, though

    ---

    - 2/10/26 found that was generating sequence incorrectly (needed to resample actions rather than shuffle them), and now all the above seems to be fixed! for now, reverting back to use flat=False to make things easier for model as a starting point. but still using variational HMM because there is some weird order effect when using basic categorical model
    """)
    return


@app.class_definition
class DebuggingMonitor(ConvergenceMonitor):
    class PROPS():
        NORM_POSTERIORS = ['startprob_', 'transmat_', 'emissionprob_']

    def __init__(self, model, tol, n_iter, verbose, save_props_while_fitting=[]):
        super().__init__(tol, n_iter, verbose)

        self.model: hmmlearn.vhmm.VariationalCategoricalHMM = model

        self.save_props_while_fitting_ = save_props_while_fitting
        self.model_history = []

    def report(self, log_prob):
        if len(self.save_props_while_fitting_) > 0:
            self.model_history.append(self._get_weights_record())

        super().report(log_prob)

    def _get_weights_record(self):
        return {
            name: getattr(self.model, name)
            for name in self.save_props_while_fitting_
        }


@app.class_definition
class WindowableHMM:
    """HMM that can use just batches of observations (if flat=False)."""

    def __init__(self, n_states=24, max_iter=150, flat=False,
                 save_props_while_fitting=[], custom_model_args={}):
        self.n_states = n_states
        self.token_to_idx = {t: i for i, t in enumerate(TOKENS)}

        self.model = hmmlearn.vhmm.VariationalCategoricalHMM(
            n_components=self.n_states,
            n_features=len(TOKENS),
            n_iter=max_iter,
            random_state=SEED,
            **custom_model_args
        )
        cur_monitor = self.model.monitor_
        self.model.monitor_ = DebuggingMonitor(self.model,
                                  cur_monitor.tol,
                                  cur_monitor.n_iter,
                                  cur_monitor.verbose,
                                  save_props_while_fitting = save_props_while_fitting)

        self.history = []
        self.flat = flat

    def _tokens_to_obs(self, tokens):
        return np.array([[self.token_to_idx[t]] for t in tokens])

    def add_to_history(self, tokens):
        obs = self._tokens_to_obs(tokens)
        self.history.append(obs)

    def fit(self):
        """Fit HMM to tokens"""
        sequences = np.array(self.history).reshape(-1, 1)

        if self.flat:
            self.model.fit(sequences)
        else:
            lengths = [len(seq) for seq in self.history]
            self.model.fit(sequences, lengths)

    def predict(self, tokens):
        """Get state distribution after input observations (accepts a single window)"""
        try:
            obs = self._tokens_to_obs(tokens)
            _, posteriors = self.model.score_samples(obs, [len(obs)])
            state_dist = posteriors[-1]

            # Compute next state prob dist and emission prob dist
            next_state_dist = state_dist @ self.model.transmat_
            pred_dist = next_state_dist @ self.model.emissionprob_
        except Exception as e:
            print(e)
            pred_dist = np.full(len(TOKENS), np.nan)

        pred_pretty = {t: pred_dist[i] for i, t in enumerate(TOKENS)}

        frequentist_pred = {
            next_token: self._get_conditional_frequency(tokens, next_token)
            for next_token in TOKENS
        }

        return {"hmm": pred_pretty, "frequentist": frequentist_pred}

    def score(self, tokens):
        """Compute log prob of tokens under the model (accepts a single window)"""
        obs = self._tokens_to_obs(tokens)
        return self.model.score(obs, [len(obs)])

    def accuracy(self, token_sequences, backward_learning=False):
        """
        If backward-learning:

        Probability that the model's 2 sampled tokens after token 1
        match actual tokens 2 and 3 (in any order; i.e. disjoint probability).

        If forward-learning:

        Probability that the model's 1 sampled token after tokens 2 and 3 (in either order; i.e. disjoint probability)
        matches actual token 3.
        """
        if backward_learning:
            # implementation for backward-learning
            accuracy_rates = []
            for tokens in token_sequences:
                obs = self._tokens_to_obs(tokens)
                _, posteriors = self.model.score_samples([obs[0]], [1])
                state_dist_0 = posteriors[-1]

                # Compute true joint P(O1=a, O2=b | O0) by conditioning t=2
                # state distribution on the actual emission at t=1 (don't just
                # use the product of marginals, which ignores the dependence
                # that arises from marginalizing over hidden states)
                def _joint_prob(_o1, _o2):
                    state_dist_1 = state_dist_0 @ self.model.transmat_
                    weighted = state_dist_1 * self.model.emissionprob_[:, _o1]
                    state_dist_2 = weighted @ self.model.transmat_
                    return (state_dist_2 * self.model.emissionprob_[:, _o2]).sum()

                o1, o2 = obs[1].item(), obs[2].item()
                p_actions_match = _joint_prob(o1, o2)
                p_actions_match_rev = _joint_prob(o2, o1)

                # These two events are disjoint (mutually exclusive), so
                # P(match in either order) = P(order 1) + P(order 2).
                # Guard: if o1 == o2, both orders are the same event.
                if o1 == o2:
                    p_actions_match_unordered = p_actions_match
                else:
                    p_actions_match_unordered = p_actions_match + p_actions_match_rev

                accuracy_rates.append(p_actions_match_unordered)

            return np.array(accuracy_rates)
        else:
            # implementation for forward-learning
            accuracy_rates = []
            for tokens in token_sequences:
                obs = self._tokens_to_obs(tokens)
                o_actual = obs[2].item()

                def _p_outcome_given_ordering(_a1, _a2):
                    action_obs = np.array([[_a1], [_a2]])
                    _, posteriors = self.model.score_samples(action_obs, [2])
                    state_dist = posteriors[-1]
                    next_state_dist = state_dist @ self.model.transmat_
                    p_correct = (next_state_dist * self.model.emissionprob_[:, o_actual]).sum()
                    lift_idx = self.token_to_idx["lift"]
                    fail_idx = self.token_to_idx["fail"]
                    p_outcomes = (
                        (next_state_dist * self.model.emissionprob_[:, lift_idx]).sum()
                        + (next_state_dist * self.model.emissionprob_[:, fail_idx]).sum()
                    )
                    return p_correct, p_outcomes

                a1, a2 = obs[0].item(), obs[1].item()
                p_correct_1, p_total_1 = _p_outcome_given_ordering(a1, a2)

                if a1 == a2:
                    p_accuracy = p_correct_1 / p_total_1 if p_total_1 > 0 else float('nan')
                else:
                    p_correct_2, p_total_2 = _p_outcome_given_ordering(a2, a1)
                    total_correct = p_correct_1 + p_correct_2
                    total_all = p_total_1 + p_total_2
                    p_accuracy = total_correct / total_all if total_all > 0 else float('nan')

                accuracy_rates.append(p_accuracy)

            return np.array(accuracy_rates)

    def accuracy_about_stronger_agent(self, token_sequences, backward_learning=False):
        """Probability that in the model's 2 sampled tokens after token 1, the one
        which is for the stronger agent matches the actual token 2 or 3 (whichever
        is for the stronger agent)

        If model is not backward-learning, not yet implemented.
        """
        if not backward_learning:
            raise Exception('not yet implemented for forward-learning model')

        B_ACTION_IDXS = {self.token_to_idx["B_try"], self.token_to_idx["B_wait"]}
        accuracy_rates = []
        for tokens in token_sequences:
            obs = self._tokens_to_obs(tokens)
            o1, o2 = obs[1].item(), obs[2].item()

            # Find B's actual action among positions 1 and 2
            if o1 in B_ACTION_IDXS:
                b = o1
            elif o2 in B_ACTION_IDXS:
                b = o2
            else:
                # B's action was at position 0 (the conditioning token);
                # nothing to predict about B
                accuracy_rates.append(np.nan)
                continue

            _, posteriors = self.model.score_samples([obs[0]], [1])
            state_dist_0 = posteriors[-1]

            # P(O1=b | O0)
            state_dist_1 = state_dist_0 @ self.model.transmat_
            p_o1_is_b = (state_dist_1 * self.model.emissionprob_[:, b]).sum()

            # P(O2=b | O0) — marginalizing over all O1
            state_dist_2 = state_dist_1 @ self.model.transmat_
            p_o2_is_b = (state_dist_2 * self.model.emissionprob_[:, b]).sum()

            # P(O1=b AND O2=b | O0)
            weighted = state_dist_1 * self.model.emissionprob_[:, b]
            state_dist_2_cond = weighted @ self.model.transmat_
            p_both_b = (state_dist_2_cond * self.model.emissionprob_[:, b]).sum()

            # Inclusion-exclusion
            p_any_b = p_o1_is_b + p_o2_is_b - p_both_b
            accuracy_rates.append(p_any_b)

        return np.array(accuracy_rates)

    def _get_conditional_frequency(self, conditional_tokens, next_token):
        n = len(conditional_tokens) + 1
        target_ngram = tuple(conditional_tokens) + (next_token,)
        context = tuple(conditional_tokens)
        match_count = 0
        context_count = 0
        if self.flat:
            # treat all windows as one continuous sequence
            all_tokens = [
                TOKENS[idx[0]] for window_obs in self.history for idx in window_obs
            ]
            for i in range(len(all_tokens) - n + 1):
                ngram = tuple(all_tokens[i : i + n])
                if ngram[: len(context)] == context:
                    context_count += 1
                    if ngram == target_ngram:
                        match_count += 1
        else:
            # treat each window independently, only checking the start of each window
            for window_obs in self.history:
                window_tokens = [TOKENS[idx[0]] for idx in window_obs]
                if (
                    len(window_tokens) >= n
                    and tuple(window_tokens[: len(context)]) == context
                ):
                    context_count += 1
                    if tuple(window_tokens[:n]) == target_ngram:
                        match_count += 1
        return match_count / context_count if context_count > 0 else np.nan

    @property
    def convergence_history(self):
        """Returns history of log likelihoods"""
        return np.array(self.model.monitor_.history)

    @property
    def fitting_history(self):
        """Returns history of hidden weights"""
        return np.array(self.model.monitor_.model_history)


@app.cell
def _():
    np.random.seed(SEED)
    pair = AgentPair(discriminability=0.5, discriminability_noise=0.5)
    windows = generate_windows(pair, num_windows=500)
    results = fit_predict_model(windows)
    return pair, results, windows


@app.function(hide_code=True)
def visualize_agent_contributions_sampdist(pair):
    # Simulate sampling distribution of A and B, then visualize
    _n_samples = 10000
    np.random.seed(SEED)

    _A_contributions = np.array(
        [pair.A_competence * pair.A_effort() for _ in range(_n_samples)]
    )
    _B_contributions = np.array(
        [pair.B_competence * pair.B_effort() for _ in range(_n_samples)]
    )
    _combined = _A_contributions + _B_contributions

    if np.max(_A_contributions) > np.max(_B_contributions):
        print(Warning('⚠️ Assumption violated! A is currently able to contribute more than B. B should always be the stronger agent, by convention'))

    _threshold = 5

    _fig, _axes = plt.subplots(1, 3, figsize=(14, 4))

    _axes[0].hist(_A_contributions, bins=50, color="#e67e22", alpha=0.7, density=True)
    _axes[0].axvline(
        _threshold, color="black", linestyle="--", label=f"threshold={_threshold}"
    )
    _axes[0].set_title(f"A contributions (competence={pair.A_competence:.1f})")
    _axes[0].set_xlabel("Contribution")
    _axes[0].set_xlim(0, 10)
    _pA = np.mean(_A_contributions >= _threshold)
    _axes[0].legend([f"P(A>={_threshold}) = {_pA:.2f}"])

    _axes[1].hist(_B_contributions, bins=50, color="#3498db", alpha=0.7, density=True)
    _axes[1].axvline(
        _threshold, color="black", linestyle="--", label=f"threshold={_threshold}"
    )
    _axes[1].set_title(f"B contributions (competence={pair.B_competence:.1f})")
    _axes[1].set_xlabel("Contribution")
    _axes[1].set_xlim(0, 10)
    _pB = np.mean(_B_contributions >= _threshold)
    _axes[1].legend([f"P(B>={_threshold}) = {_pB:.2f}"])

    _axes[2].hist(
        _A_contributions, bins=50, color="#e67e22", alpha=0.5, density=True, label="A"
    )
    _axes[2].hist(
        _B_contributions, bins=50, color="#3498db", alpha=0.5, density=True, label="B"
    )
    _axes[2].hist(
        _combined, bins=50, color="#2ecc71", alpha=0.5, density=True, label="A+B"
    )
    _axes[2].axvline(_threshold, color="black", linestyle="--")
    _pAB = np.mean(_combined >= _threshold)
    _axes[2].set_title(f"Combined: P(A+B>={_threshold}) = {_pAB:.2f}")
    _axes[2].set_xlabel("Contribution")
    _axes[2].legend()

    plt.suptitle(
        f"discriminability={pair.discriminability}, noise={pair.discriminability_noise}, A competence={pair.A_competence}",
        fontweight="bold",
    )
    plt.tight_layout()
    return _fig


@app.cell
def _(pair):
    visualize_agent_contributions_sampdist(pair)
    return


@app.cell(hide_code=True)
def _(results):
    mo.vstack(
        [
            plot_results(results),
            mo.md("""
        ~~**current interpretation of the double-descent in both models (when `flat=True`):** HMM starts learning statistical patterns from the start, but then after enough iterations it learns that, in forward model: the outcome does not predict the next token (first action of another sequence) at all; or in backward model: each second action token does not predict the next token (outcome of another sequence) at all. this was inferred based on the double-descent only appearing when fitting the HMM on continuous strings of tokens instead of discrete sequences (i.e. as separate windows)~~

        **one way to understand double-descent even when `flat=False` (happens sometime, at least with variational categorical HMM):** HMM first learns a spurious relationship between agents' actions/maybe cares about their order, but then learns that it's irrelevant and benefits from discarding/deprioritizing this information
              """),
        ]
    )
    return


@app.cell
def _(results):
    mo.vstack(
        [
            # before fitting
            plot_hmm_matrices(
                results["model_forward"].model.startprob_prior_.reshape(
                    results["model_forward"].n_states, 1
                ),
                results["model_forward"].model.transmat_prior_,
                results["model_forward"].model.emissionprob_prior_,
            ),
            # ???
            plot_hmm_matrices(
                results["model_forward"].model.startprob_.reshape(
                    results["model_forward"].n_states, 1
                ),
                results["model_forward"].model.transmat_,
                results["model_forward"].model.emissionprob_,
            ),
            # after fitting
            plot_hmm_matrices(
                results["model_forward"].model.startprob_posterior_.reshape(
                    results["model_forward"].n_states, 1
                ),
                results["model_forward"].model.transmat_posterior_,
                results["model_forward"].model.emissionprob_posterior_,
            ),
        ]
    )
    return


@app.cell
def _(results):
    for _i in range(50):
        _observed, _hidden = results["model_forward"].model.sample(3, random_state=_i)
        print([TOKENS[_i] for _i in _observed.reshape(-1)])

    # can only keep going if flat=True
    # _observed, _hidden = results['model_forward'].model.sample(3, currstate=_hidden[-1])
    # print([TOKENS[i] for i in _observed.reshape(-1)])
    return


@app.cell
def _(results):
    mo.vstack(
        [
            # before fitting
            plot_hmm_matrices(
                results["model_backward"].model.startprob_prior_.reshape(
                    results["model_backward"].n_states, 1
                ),
                results["model_backward"].model.transmat_prior_,
                results["model_backward"].model.emissionprob_prior_,
            ),
            # after fitting
            plot_hmm_matrices(
                results["model_backward"].model.startprob_.reshape(
                    results["model_backward"].n_states, 1
                ),
                results["model_backward"].model.transmat_,
                results["model_backward"].model.emissionprob_,
            ),
        ]
    )
    return


@app.cell
def _(results):
    for _i in range(50):
        _observed, _hidden = results["model_backward"].model.sample(3, random_state=_i)
        print(list(reversed([TOKENS[_i] for _i in _observed.reshape(-1)])))

    # can only keep going if flat=True
    # _observed, _hidden = results['model_backward'].model.sample(3, currstate=_hidden[-1])
    # print(list(reversed([TOKENS[i] for i in _observed.reshape(-1)])))
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## tests to evaluate learning

    ### for forward model:
    - order of 2 actions shouldn't matter
    - TT should have higher lift likelihood than of TW (iff both agents are capable on their own)
    - TW should be sensitive to who's lifting (iff one agent is strong enough but the other isn't)
    - likelihood of fail for WW should be ~1
    - for T, lift and fail should each be impossible
    - agent shouldn't be able to take two actions in a row
    - ~~shouldn't be able to have two outcomes in a row?~~ behavior is not defined here

    > IDEA: visualize these tests/explicitly perform these tests

    ### for backward model:
    - should make smart inference about what happened in the two actions which preceded a lift or fail, based on the strengths of the agents (B_competence >= A_competence by my convention, so B is the stronger agent)
    """)
    return


@app.cell(hide_code=True)
def _(results):
    results["pred"]
    return


@app.function
def invisible(arg):
    return None


@app.function
def show_results_accuracy(results, windows):
    fwd_acc = results["model_forward"].accuracy(windows)
    bwd_acc = results["model_backward"].accuracy(
        [list(reversed(w)) for w in windows],
        backward_learning=True
    )

    # fwd_acc_b = results["model_forward"].accuracy_about_stronger_agent(windows)
    bwd_acc_b = results["model_backward"].accuracy_about_stronger_agent(
        [list(reversed(w)) for w in windows],
        backward_learning=True
    )

    acc_results = pl.concat([
        pl.DataFrame({
            'Forward': fwd_acc.reshape(-1),
            'Backward': bwd_acc.reshape(-1),
        })
        .unpivot(variable_name='model', value_name='accuracy')
        .with_columns(target=pl.lit('both agents')),

        pl.DataFrame({
            # 'Forward': fwd_acc_b.reshape(-1),
            'Backward': bwd_acc_b.reshape(-1),
        })
        .unpivot(variable_name='model', value_name='accuracy')
        .with_columns(target=pl.lit('stronger agent'))
    ])

    return mo.vstack([
        mo.hstack([
            mo.md(f"""
            accurate = sampled tokens 2 & 3 match actual, any order:  
            Forward model: {fwd_acc.mean():.1%}, Backward model: {bwd_acc.mean():.1%}
            """),
            mo.md(f"""
            accurate = matches stronger agent's token:  
            {invisible("Forward model: {np.nanmean(fwd_acc_b):.1%}")}, Backward model: {np.nanmean(bwd_acc_b):.1%}
            """)
        ]),
        sns.catplot(
            acc_results.to_pandas(),
            kind='point',
            x='model',
            y='accuracy',
            col='target'
        ).map_dataframe(
            sns.stripplot,
            x='model',
            y='accuracy',
            alpha=.1
        )
    ])


@app.cell
def _(pair, results, windows):
    mo.vstack([
        mo.md("**Within-sample accuracy**"),
        show_results_accuracy(results, windows),
        mo.md("**Out-of-sample accuracy**"),
        show_results_accuracy(results, generate_windows(pair, num_windows=500)),
    ])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Comparison to null model
    """)
    return


@app.cell
def _(windows):
    null_windows = shuffle(np.array(windows).reshape(-1)).reshape(len(windows), 3)
    # null model (same model fit to windows without consistent structure)
    results_null = fit_predict_model(null_windows, max_iter=100)
    return null_windows, results_null


@app.cell(hide_code=True)
def _(results_null):
    plot_results(results_null)
    return


@app.cell(hide_code=True)
def _(results_null):
    mo.vstack(
        [
            # before fitting
            plot_hmm_matrices(
                results_null["model_forward"].model.startprob_prior_.reshape(
                    results_null["model_forward"].n_states, 1
                ),
                results_null["model_forward"].model.transmat_prior_,
                results_null["model_forward"].model.emissionprob_prior_,
            ),
            # after fitting
            plot_hmm_matrices(
                results_null["model_forward"].model.startprob_.reshape(
                    results_null["model_forward"].n_states, 1
                ),
                results_null["model_forward"].model.transmat_,
                results_null["model_forward"].model.emissionprob_,
            ),
            # before fitting
            plot_hmm_matrices(
                results_null["model_backward"].model.startprob_prior_.reshape(
                    results_null["model_backward"].n_states, 1
                ),
                results_null["model_backward"].model.transmat_prior_,
                results_null["model_backward"].model.emissionprob_prior_,
            ),
            # after fitting
            plot_hmm_matrices(
                results_null["model_backward"].model.startprob_.reshape(
                    results_null["model_backward"].n_states, 1
                ),
                results_null["model_backward"].model.transmat_,
                results_null["model_backward"].model.emissionprob_,
            ),
        ]
    )
    return


@app.cell(hide_code=True)
def _(null_windows, pair, results_null, windows):
    _new_null_windows = shuffle(np.array(null_windows).reshape(-1)).reshape(len(windows), 3)
    mo.vstack([
        mo.md("**_Null model_ Within-sample accuracy (null windows)**"),
        show_results_accuracy(results_null, null_windows),
        mo.md("**_Null model_ Out-of-sample accuracy (null windows)**"),
        show_results_accuracy(results_null, _new_null_windows),
        mo.md("**_Null model_ Out-of-sample accuracy (real windows)**"),
        show_results_accuracy(results_null, generate_windows(pair, num_windows=500))
    ])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Sequence generation & model visualization code
    """)
    return


@app.function
def shuffle(arr):
    return np.random.choice(arr, size=np.shape(arr), replace=False)


@app.function
def resample(arr):
    return np.random.choice(arr, size=np.shape(arr), replace=True)


@app.function
def generate_outcome(pair, action1, action2):
    competences = {
        "A_try": pair.A_competence,  # A's competence
        "B_try": pair.B_competence,  # B's competence
        "A_wait": 0,
        "B_wait": 0,
    }
    efforts = {
        "A_try": pair.A_effort(),  # A's effort
        "B_try": pair.B_effort(),  # B's effort
        "A_wait": 0,  # doesn't matter since multiplying 0
        "B_wait": 0,  # doesn't matter since multiplying 0
    }
    action1_contribution = competences[action1] * efforts[action1]
    if action1 == action2:
        combined_contribution = action1_contribution
    else:
        action2_contribution = competences[action2] * efforts[action2]
        combined_contribution = action1_contribution + action2_contribution

    minimum_contribution = 5  # weight of 5
    success = combined_contribution >= minimum_contribution
    outcome = "lift" if success else "fail"
    return outcome


@app.function
def generate_windows(pair, num_windows):
    windows = []
    for _ in range(num_windows):
        agent1, agent2 = shuffle(["A", "B"])
        agent1_action, agent2_action = resample(["try", "wait"])
        action1 = f"{agent1}_{agent1_action}"
        action2 = f"{agent2}_{agent2_action}"
        outcome = generate_outcome(pair, action1, action2)
        current_window = [action1, action2, outcome]
        windows.append(current_window)
    return windows


@app.function
def make_null_actions_generator():
    """
    For generating random windows where the transitions between actions are random.
    Useful to test test whether the HMM needs to learn temporal contingencies between actions (e.g., that each agent takes exactly one action, and they are unrelated) to predict the outcome.
    """
    # The first two tokens in regular windows can be interpreted as a two-state sequence from a Markov chain:
    # Start -(Uniform probability)-> A_try, A_wait, B_try, B_wait
    # A_try -(50%)-> B_try
    # A_try -(50%)-> B_wait
    # B_try -(50%)-> A_try
    # B_try -(50%)-> A_wait
    # A_wait -(50%)-> B_wait
    # A_wait -(50%)-> B_try
    # B_wait -(50%)-> A_wait
    # B_wait -(50%)-> A_try
    # i.e. transition matrix is 4x4:
    # A_try A_wait B_try B_wait
    # 0  0   .5 .5
    # 0  0   .5 .5
    # .5 .5  0  0
    # .5 .5  0  0
    # The first two tokens in windows will null actions can be interpreted as a state sequence from a two-state Markov chain:
    # the transition matrix is 4x4 but randomized ~~as though it's an Erdos-Renyi random graph with p=(8/16) and weight=unif(0,1)~~
    # should it be matched on more dimensions? ideally, it should be matched on all dimensions except the one I'm testing.
    # so let's have it matched in transitions being uniform (unbiased), and also have the matrix still be row-stochastic (so that an action always leads to a second action; i.e. so that the sequence never terminates)
    # ~~and let's also have it match in having a stationary uniform distribution~~ this is actually hard, so overkill rn.

    # To do this: for each row of the matrix, select n in 1,2,3,4 and randomly select n cells in the row.
    # Then fill each selected cell with 1/n (as the transition probability)
    trans_mat = np.zeros((4, 4))
    for i in range(4):
        num_indices = np.random.randint(0, 4) + 1
        indices = np.random.choice(np.arange(4), size=num_indices, replace=False)
        trans_mat[i, indices] = 1./num_indices

    start_prob = np.ones(4) / 4

    def _generate(max_length=2):
        trans_prob = start_prob
        for _ in range(max_length):
            state = np.random.choice(np.arange(trans_prob.shape[0]), p=trans_prob)
            yield state
            trans_prob = trans_mat[state]

    return _generate, start_prob, trans_mat


@app.function
def generate_windows_with_actions_generator(pair, num_windows, generator):
    # The outcome is still dependent on the actions.
    # ~~I think I'll still allow double-counting actions to keep constant the outcome-decoding process (as a function of the actions).~~
    # Actually, we get a confound if we allow double-counting-- we just get more diverse training data, which on its own may make it easier for the HMM to learn the generate_outcome function.
    # So I modified generate_outcome to reflect this; our target generate_outcome function need not be designed to allow double-counting, anyway.
    windows = []
    for _ in range(num_windows):
        actions = list(generator(max_length=2))
        action1 = TOKENS[actions[0]]
        action2 = TOKENS[actions[1]]
        outcome = generate_outcome(pair, action1, action2)
        current_window = [action1, action2, outcome]
        windows.append(current_window)
    return windows


@app.cell(disabled=True)
def _(pair):
    _gen, _start_prob, _trans_mat = make_null_actions_generator()
    _null_windows = generate_windows_with_actions_generator(pair, 1000, _gen)

    # validating that the startprob frequences are good
    _first_states = []
    for _win in _null_windows:
        _first_states.append(_win[0])

    # validating that the ngram frequences match the transition frequencies
    _ngrams = []
    for _win in _null_windows:
        _ngrams.append(f"{_win[0]}->{_win[1]}")

    print(_start_prob)
    sns.barplot(_first_states)
    print(_trans_mat)
    sns.barplot(_ngrams)
    return


@app.function
def fit_predict_model(windows, n_states=24, max_iter=100):
    hmm_model_forward = WindowableHMM(n_states=n_states, max_iter=max_iter)
    hmm_model_backward = WindowableHMM(n_states=n_states, max_iter=max_iter)

    for window in windows:
        hmm_model_forward.add_to_history(window)
        hmm_model_backward.add_to_history(list(reversed(window)))

    hmm_model_forward.fit()
    hmm_model_backward.fit()

    pred_forward = {}

    # consider both orderings (should be the same in this case, since HMM should not learn that there's any interdependence/importance of order)
    pred_forward["TT?"] = {
        "A_try,B_try,?": hmm_model_forward.predict(["A_try", "B_try"]),
        "B_try,A_try,?": hmm_model_forward.predict(["B_try", "A_try"]),
    }

    pred_forward["TW?"] = {
        "A_try,B_wait,?": hmm_model_forward.predict(["A_try", "B_wait"]),
        "B_try,A_wait,?": hmm_model_forward.predict(["B_try", "A_wait"]),
        "B_wait,A_try,?": hmm_model_forward.predict(["B_wait", "A_try"]),
        "A_wait,B_try,?": hmm_model_forward.predict(["A_wait", "B_try"]),
    }

    pred_forward["WW?"] = {
        "A_wait,B_wait,?": hmm_model_forward.predict(["A_wait", "B_wait"]),
        "B_wait,A_wait,?": hmm_model_forward.predict(["B_wait", "A_wait"]),
    }

    pred_forward["T?"] = {
        "A_try,?": hmm_model_forward.predict(["A_try"]),
        "B_try,?": hmm_model_forward.predict(["B_try"]),
    }

    pred_backward = {}

    pred_backward["?Outcome"] = {
        "?,lift": hmm_model_backward.predict(["lift"]),
        "?,fail": hmm_model_backward.predict(["fail"]),
    }

    pred_backward["?TOutcome"] = {
        "?,A_try,lift": hmm_model_backward.predict(["lift", "A_try"]),
        "?,B_try,lift": hmm_model_backward.predict(["lift", "B_try"]),
        "?,A_try,fail": hmm_model_backward.predict(["fail", "A_try"]),
        "?,B_try,fail": hmm_model_backward.predict(["fail", "B_try"]),
    }
    pred_backward["?WOutcome"] = {
        "?,A_wait,lift": hmm_model_backward.predict(["lift", "A_wait"]),
        "?,B_wait,lift": hmm_model_backward.predict(["lift", "B_wait"]),
        "?,A_wait,fail": hmm_model_backward.predict(["fail", "A_wait"]),
        "?,B_wait,fail": hmm_model_backward.predict(["fail", "B_wait"]),
    }

    return {
        "windows": windows,
        "pred": {
            "pred_forward": pred_forward,
            "pred_backward": pred_backward,
        },
        "model_forward": hmm_model_forward,
        "model_backward": hmm_model_backward,
    }


@app.function
def plot_results(results):
    tokens = np.array(results["windows"]).reshape(-1)

    fig, axes = plt.subplots(
        3,
        1,
        figsize=(12, 7),
    )

    # Panel 1: Token sequence
    ax = axes[0]
    for i, token in enumerate(tokens):
        color = TOKEN_COLORS.get(token, "#bdc3c7")
        ax.barh(0, 1, left=i, color=color, edgecolor="none")
    ax.set_xlim(0, len(tokens))
    ax.set_ylim(-0.5, 0.5)
    ax.set_yticks([])
    ax.set_xlabel("Token Index")
    ax.set_title("Observed Token Sequence", fontsize=11, fontweight="bold")
    patches = [mpatches.Patch(color=c, label=t) for t, c in TOKEN_COLORS.items()]
    ax.legend(handles=patches, loc="upper right", ncol=5, fontsize=8)

    # Panels 2-3: LogProb
    ax = axes[1]
    ax.plot(
        -results["model_forward"].convergence_history[1:],
        marker="o",
        linestyle="-",
        color="#3498db",
    )
    ax.set_title(
        "Forward Model Fit",
        fontsize=11,
        fontweight="bold",
    )
    ax.set_xlabel("Iteration")
    ax.set_ylabel("-logP")
    ax.set_xlim(0, 100)

    ax = axes[2]
    ax.plot(
        -results["model_backward"].convergence_history[1:],
        marker="o",
        linestyle="-",
        color="#3498db",
    )
    ax.set_title(
        "Backward Model Fit",
        fontsize=11,
        fontweight="bold",
    )
    ax.set_xlabel("Iteration")
    ax.set_ylabel("-logP")
    ax.set_xlim(0, 100)

    plt.tight_layout()

    return fig


@app.function
def plot_hmm_matrices(start_prob, trans_mat, emission_prob, order=None):
    _fig, _ax = plt.subplots(1, 3, figsize=(9, 3))

    assert np.shape(start_prob)[0] == np.shape(trans_mat)[0]
    assert np.shape(start_prob)[0] == np.shape(emission_prob)[0]
    if order is None:
        order = np.arange(0, np.shape(start_prob)[0])

    sns.heatmap(start_prob[order], vmin=0, vmax=1, cbar=False, yticklabels=order, xticklabels=False, ax=_ax[0])
    sns.heatmap(trans_mat[order], vmin=0, vmax=1, cbar=False, yticklabels=order, ax=_ax[1])
    sns.heatmap(emission_prob[order], vmin=0, vmax=1, yticklabels=order, xticklabels=TOKENS, ax=_ax[2])
    _fig.tight_layout()
    return _fig


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ---
    # Reversible HMM attempt (2/10/26)
    """)
    return


@app.cell
def _(SEED2):
    import deeptime

    class ReversibleHMM:
        """HMM that has a reversible transition matrix. Currently only models observations as one continuous sequence."""

        def __init__(self, n_states=20, flat=True):
            self.n_states = n_states
            self.token_to_idx = {t: i for i, t in enumerate(TOKENS)}
            self.history = []

            self.flat = flat

        def _tokens_to_obs(self, tokens):
            return np.array([self.token_to_idx[t] for t in tokens])

        def add_to_history(self, tokens):
            obs = self._tokens_to_obs(tokens)
            self.history.append(obs)

        def fit(self):
            if self.flat:
                sequences = np.array(self.history).reshape(-1)
            else:
                # sequences = np.array(self.history)
                raise Exception("flat=False not implemented")

            self.initial_model = deeptime.markov.hmm.init.discrete.random_guess(
                n_observation_states=len(TOKENS),
                n_hidden_states=self.n_states,
                seed=SEED2,
            )

            # Fit HMM to tokens
            self.model = deeptime.markov.hmm.MaximumLikelihoodHMM(
                self.initial_model, lagtime=1, reversible=True
            ).fit_fetch(sequences)

        def predict(self, tokens):
            # Get state distribution after input observations
            raise Exception("Not implemented")

        def _get_conditional_frequency(self, conditional_tokens, next_token):
            n = len(conditional_tokens) + 1
            target_ngram = tuple(conditional_tokens) + (next_token,)
            context = tuple(conditional_tokens)
            match_count = 0
            context_count = 0
            if self.flat:
                # treat all windows as one continuous sequence
                all_tokens = [
                    TOKENS[idx[0]] for window_obs in self.history for idx in window_obs
                ]
                for i in range(len(all_tokens) - n + 1):
                    ngram = tuple(all_tokens[i : i + n])
                    if ngram[: len(context)] == context:
                        context_count += 1
                        if ngram == target_ngram:
                            match_count += 1
            else:
                # treat each window independently, only checking the start of each window
                for window_obs in self.history:
                    window_tokens = [TOKENS[idx[0]] for idx in window_obs]
                    if (
                        len(window_tokens) >= n
                        and tuple(window_tokens[: len(context)]) == context
                    ):
                        context_count += 1
                        if tuple(window_tokens[:n]) == target_ngram:
                            match_count += 1
            return match_count / context_count if context_count > 0 else np.nan

        @property
        def convergence_history(self):
            return np.array(self.model.likelihoods)

    return (ReversibleHMM,)


@app.cell
def _():
    SEED2 = 6
    return (SEED2,)


@app.cell
def _(ReversibleHMM, SEED2, windows):
    np.random.seed(SEED2)
    hmm_model = ReversibleHMM(flat=True)

    for window in windows:
        hmm_model.add_to_history(window)

    hmm_model.fit()

    plt.plot(-hmm_model.convergence_history)
    return (hmm_model,)


@app.cell
def _(hmm_model):
    mo.vstack(
        [
            # before fitting
            plot_hmm_matrices(
                hmm_model.initial_model.initial_distribution.reshape(
                    hmm_model.n_states, 1
                ),
                hmm_model.initial_model.transition_model.transition_matrix,
                hmm_model.initial_model.output_probabilities,
            ),
            # after fitting
            plot_hmm_matrices(
                hmm_model.model.initial_distribution.reshape(hmm_model.n_states, 1),
                hmm_model.model.transition_model.transition_matrix,
                hmm_model.model.output_probabilities,
            ),
        ]
    )
    return


@app.cell
def _(hmm_model):
    _hidden, _observed = hmm_model.model.simulate(3)
    print([TOKENS[i] for i in _observed])

    _hidden, _observed = hmm_model.model.simulate(3, start=_hidden[-1])
    print([TOKENS[i] for i in _observed])

    _hidden, _observed = hmm_model.model.simulate(3, start=_hidden[-1])
    print([TOKENS[i] for i in _observed])

    _hidden, _observed = hmm_model.model.simulate(3, start=_hidden[-1])
    print([TOKENS[i] for i in _observed])

    _hidden, _observed = hmm_model.model.simulate(3, start=_hidden[-1])
    print([TOKENS[i] for i in _observed])
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "marimo>=0.19.0",
#     "pyzmq",
# ]
# ///

# NOTE: Use the command palette (cmd+shift+p) and search "marimo start in editor" to generate the live preview.

import marimo

__generated_with = "0.13.6"
app = marimo.App()


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    # Marimo Notebook Demo

    Marimo Notebooks are a newer tool that combine the best of scripts, Python notebooks, and Quarto notebooks/documents. Marimo notebooks have their own mini user-interface rather than embedding in Cursor/VSCode.

    But, unlike normal Python notebooks everything is all stored in a `.py` file which you can run like any other Python file e.g. `pixi run python yourmarimofile.py`.

    We've configured this project to let you start marimo and view it in a seperate browser if you prefer using

    `pixi run marimo file.py`

    Or you can use the VSCode extension but using the command palette (`cmd+shift+p`) and searching "marimo start in editor", which will open the marimo notebook preview in VSCode itself.
    """
    )
    return


@app.cell
def _():
    # Same cell as our Python notebook
    # This is a custom function we wrote save in src/lib/helpers.py
    from lib.helpers import hello_world

    hello_world()
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    Cells are also *reactive* by default which means they track all the values that variables and code depend on. In this next cell, we create a radio-button and save it to the `radiogroup` variable.

    Then we can read the `radiogroup.value` attribute in the any code-cell below that and it will always auto-update to the latest value of which option was selected!
    """
    )
    return


@app.cell(hide_code=True)
def _():
    import marimo as mo

    radiogroup = mo.ui.radio(
        options={"one": 1, "two": 2, "three": 3},
        value="one",
        label="pick a number",
    )
    radiogroup
    return mo, radiogroup


@app.cell
def _(radiogroup):
    print("Watch this value react!")
    print(f"The value is {radiogroup.value}")
    return


@app.cell(hide_code=True)
def _():
    # Import some libraries
    import numpy as np
    import seaborn as sns
    import polars as pl

    # Set the path for saving figures
    from pathlib import Path

    figures_path = Path("./figures/")
    return np, pl, sns


@app.cell
def _(np, pl, radiogroup, sns):
    # Generate some random data
    x = np.random.randn(100)
    y = np.random.randn(100) * radiogroup.value
    df = pl.DataFrame({"x": x, "y": y})

    # Plot the data
    grid = sns.FacetGrid(df)
    grid.map(sns.scatterplot, "x", "y")

    # Customize the plot
    grid.set(
        title="I saved this plot to figures/demo.png!",
        xlim=(-3, 3),
        ylim=(-3, 3),
        xlabel="Some random numbers",
        ylabel="Some other random numbers",
    )
    return (df,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Dataframes render interactively with mini-histograms""")
    return


@app.cell(hide_code=True)
def _(df, pl):
    df.select(pl.col("x").abs())
    return


@app.cell
def _(mo):
    mo.md(r""" """)
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "marimo>=0.19.0",
#     "pyzmq",
# ]
# ///

# NOTE: Use the command palette (cmd+shift+p) and search "marimo start in editor" to generate the live preview.

import marimo

__generated_with = "0.19.6"
app = marimo.App()


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Compare sliding-window HMM vs online Bayesian model for trait estimation.

    Focus: How do trait estimates (warmth, competence, effort) evolve over time?

    - Bayesian: Accumulates all evidence, converges to global average
    - HMM: Uses recent window, tracks local behavioral "mode"
    """)
    return


@app.cell(hide_code=True)
def _():
    import marimo as mo
    import numpy as np
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches
    from collections import Counter, defaultdict
    import warnings
    warnings.filterwarnings('ignore')

    from hmmlearn import hmm
    import networkx as nx
    return Counter, defaultdict, hmm, mo, mpatches, np, nx, plt


@app.cell(hide_code=True)
def _():
    TOKENS = ["try_success", "try_fail", "help", "harm", "wait"]
    TOKEN_COLORS = {
        "try_success": "#2ecc71",
        "try_fail": "#e74c3c", 
        "help": "#3498db",
        "harm": "#9b59b6",
        "wait": "#95a5a6"
    }

    TRAIT_COLORS = {
        "effort": "#e67e22",
        "competence": "#27ae60", 
        "warmth": "#2980b9"
    }

    mode_colors = {
        "prosocial": "violet",
        "antisocial": "orange", 
        "withdrawn": "black"
    }

    outcome_colors = {
        "success": "#2ecc71",
        "fail": "#e74c3c",
        "help": "#3498db",
        "harm": "#9b59b6",
        "wait": "#95a5a6",
        "try": "#e67e22"
    }
    return TOKENS, TOKEN_COLORS, mode_colors, outcome_colors


@app.cell(hide_code=True)
def _(np):
    from dataclasses import dataclass, field
    from abc import ABC, abstractmethod

    @dataclass
    class BaseAgent(ABC):
        """Base class for behavioral agents with mode-based action generation."""
        modes: dict = field(default_factory=dict)
        competence: dict = field(default_factory=dict)
        transitions: dict = field(default_factory=dict)
        current_mode: str = "prosocial"
        initial_mode: str = "prosocial"

        def reset(self):
            self.current_mode = self.initial_mode

        def generate_token(self):
            probs = self.modes[self.current_mode]
            actions = list(probs.keys())
            p = list(probs.values())
            action = np.random.choice(actions, p=p)

            if action == "try":
                comp = self.competence[self.current_mode]
                outcome = "success" if np.random.random() < comp else "fail"
                token = f"try_{outcome}"
                self._transition(outcome)
            else:
                token = action
                self._transition(action)

            return token, self.current_mode

        def _transition(self, outcome):
            key = (self.current_mode, outcome)
            if key in self.transitions:
                trans = self.transitions[key]
                modes = list(trans.keys())
                probs = list(trans.values())
                self.current_mode = np.random.choice(modes, p=probs)


    @dataclass
    class VolatileAgent(BaseAgent):
        """Agent with distinct prosocial/antisocial/withdrawn phases."""

        def __post_init__(self):
            self.modes = {
                "prosocial": {"try": 0.4, "help": 0.5, "harm": 0.0, "wait": 0.1},
                "antisocial": {"try": 0.2, "help": 0.0, "harm": 0.6, "wait": 0.2},
                "withdrawn": {"try": 0.1, "help": 0.1, "harm": 0.1, "wait": 0.7},
            }
            self.competence = {"prosocial": 0.7, "antisocial": 0.7, "withdrawn": 0.4}
            self.transitions = {
                ("prosocial", "success"): {"prosocial": 0.85, "antisocial": 0.05, "withdrawn": 0.1},
                ("prosocial", "fail"): {"prosocial": 0.6, "antisocial": 0.2, "withdrawn": 0.2},
                ("prosocial", "help"): {"prosocial": 0.9, "antisocial": 0.0, "withdrawn": 0.1},
                ("prosocial", "harm"): {"prosocial": 0.35, "antisocial": 0.45, "withdrawn": 0.2},
                ("prosocial", "wait"): {"prosocial": 0.6, "antisocial": 0.1, "withdrawn": 0.3},
                ("antisocial", "success"): {"prosocial": 0.3, "antisocial": 0.6, "withdrawn": 0.1},
                ("antisocial", "fail"): {"prosocial": 0.1, "antisocial": 0.7, "withdrawn": 0.2},
                ("antisocial", "help"): {"prosocial": 0.35, "antisocial": 0.45, "withdrawn": 0.2},
                ("antisocial", "harm"): {"prosocial": 0.0, "antisocial": 0.9, "withdrawn": 0.1},
                ("antisocial", "wait"): {"prosocial": 0.2, "antisocial": 0.5, "withdrawn": 0.3},
                ("withdrawn", "success"): {"prosocial": 0.35, "antisocial": 0.35, "withdrawn": 0.3},
                ("withdrawn", "fail"): {"prosocial": 0.2, "antisocial": 0.2, "withdrawn": 0.6},
                ("withdrawn", "help"): {"prosocial": 0.5, "antisocial": 0.2, "withdrawn": 0.3},
                ("withdrawn", "harm"): {"prosocial": 0.1, "antisocial": 0.5, "withdrawn": 0.4},
                ("withdrawn", "wait"): {"prosocial": 0.2, "antisocial": 0.2, "withdrawn": 0.6},
            }
            self.current_mode = "prosocial"
            self.initial_mode = "prosocial"


    @dataclass
    class LimitedEffortAgent(BaseAgent):
        """Agent that is almost always prosocial but has decreasing effort over time."""
        initial_effort: float = 0.9
        effort_decay: float = 0.01
        min_effort: float = 0.1
        current_effort: float = field(init=False)
        step: int = field(init=False)

        def __post_init__(self):
            self.current_effort = self.initial_effort
            self.step = 0
            self.modes = {
                "prosocial": {"try": 0.5, "help": 0.4, "harm": 0.02, "wait": 0.08},
                "antisocial": {"try": 0.2, "help": 0.0, "harm": 0.6, "wait": 0.2},
                "withdrawn": {"try": 0.1, "help": 0.1, "harm": 0.1, "wait": 0.7},
            }
            self.competence = {"prosocial": 0.75, "antisocial": 0.7, "withdrawn": 0.4}
            self.transitions = {
                ("prosocial", "success"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("prosocial", "fail"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("prosocial", "help"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("prosocial", "harm"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("prosocial", "wait"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("antisocial", "success"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("antisocial", "fail"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("antisocial", "help"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("antisocial", "harm"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("antisocial", "wait"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("withdrawn", "success"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("withdrawn", "fail"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("withdrawn", "help"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("withdrawn", "harm"): {"prosocial": 1.0, "antisocial": 0.0, "withdrawn": 0.0},
                ("withdrawn", "wait"): {"prosocial": .2, "antisocial": 0.0, "withdrawn": 0.8},
            }
            self.current_mode = "prosocial"
            self.initial_mode = "prosocial"

        def reset(self):
            self.current_effort = self.initial_effort
            self.step = 0
            self.current_mode = self.initial_mode

        def generate_token(self):
            self.step += 1
            self.current_effort = max(
                self.min_effort,
                self.initial_effort - self.effort_decay * self.step
            )

            if self.current_effort < .2:
                self.current_mode = "withdrawn"

            # Use base class token generation
            probs = self.modes[self.current_mode]
            actions = list(probs.keys())
            p = list(probs.values())
            action = np.random.choice(actions, p=p)

            if action == "try":
                comp = self.competence[self.current_mode]
                outcome = "success" if np.random.random() < comp else "fail"
                token = f"try_{outcome}"
            else:
                token = action

            return token, self.current_mode
    return LimitedEffortAgent, VolatileAgent


@app.cell
def _(mo):
    agent_selector = mo.ui.dropdown(
        options=["VolatileAgent", "LimitedEffortAgent"],
        value="VolatileAgent",
        label="Current agent",
        full_width=False
    )
    agent_selector
    return (agent_selector,)


@app.cell
def _(LimitedEffortAgent, VolatileAgent, agent_selector):
    if agent_selector.value == "VolatileAgent":
        Agent = VolatileAgent
    elif agent_selector.value == "LimitedEffortAgent":
        Agent = LimitedEffortAgent
    else:
        raise ValueError("Unknown agent type selected.")
    return (Agent,)


@app.class_definition(hide_code=True)
class OnlineBayesian:
    """Bayesian model with conjugate Beta priors for trait estimation."""

    def __init__(self, prior_alpha=1.0, prior_beta=1.0):
        # Effort: P(not wait)
        self.effort_alpha = prior_alpha
        self.effort_beta = prior_beta
        # Competence: P(success | try)
        self.comp_alpha = prior_alpha
        self.comp_beta = prior_beta
        # Warmth components
        self.help_alpha = prior_alpha
        self.help_beta = prior_beta
        self.harm_alpha = prior_alpha
        self.harm_beta = prior_beta

        self.n_obs = 0

    def update(self, token):
        self.n_obs += 1

        # Update effort (not wait vs wait)
        if token == "wait":
            self.effort_beta += 1
        else:
            self.effort_alpha += 1

        # Update competence (success vs fail, conditional on try)
        if token == "try_success":
            self.comp_alpha += 1
        elif token == "try_fail":
            self.comp_beta += 1

        # Update warmth components
        if token == "help":
            self.help_alpha += 1
        else:
            self.help_beta += 1

        if token == "harm":
            self.harm_alpha += 1
        else:
            self.harm_beta += 1

    def get_traits(self):
        effort = self.effort_alpha / (self.effort_alpha + self.effort_beta)
        competence = self.comp_alpha / (self.comp_alpha + self.comp_beta)
        p_help = self.help_alpha / (self.help_alpha + self.help_beta)
        p_harm = self.harm_alpha / (self.harm_alpha + self.harm_beta)
        warmth = p_help - p_harm

        return {"effort": effort, "competence": competence, "warmth": warmth}


@app.cell(hide_code=True)
def _(Counter, TOKENS, hmm, np):
    class SlidingWindowHMM:
        """HMM that uses only recent observations to estimate traits."""

        def __init__(self, window_size=20, n_states=3):
            self.window_size = window_size
            self.n_states = n_states
            self.tokens = TOKENS
            self.n_tokens = len(TOKENS)
            self.token_to_idx = {t: i for i, t in enumerate(TOKENS)}
            self.history = []

        def update(self, token):
            self.history.append(token)

        def get_traits(self):
            """Get traits from HMM trained on recent window."""
            if len(self.history) < 5:
                return {"effort": 0.5, "competence": 0.5, "warmth": 0.0}

            # Use sliding window
            window = self.history[-self.window_size:] if len(self.history) > self.window_size else self.history

            # Fit HMM to window
            obs = np.array([[self.token_to_idx[t]] for t in window])

            model = hmm.CategoricalHMM(
                n_components=self.n_states,
                n_iter=50,
                random_state=42
            )
            model.n_features = self.n_tokens

            try:
                model.fit(obs)

                # Get current state distribution
                _, posteriors = model.score_samples(obs)
                current_state = posteriors[-1]

                # Predict next state and emission
                next_state = current_state @ model.transmat_
                pred = next_state @ model.emissionprob_

            except:
                # Fallback to empirical counts from window
                counts = Counter(window)
                total = len(window)
                pred = np.array([counts.get(t, 0) / total for t in TOKENS])

            # Extract traits from prediction
            p_try = pred[0] + pred[1]  # try_success + try_fail
            p_wait = pred[4]
            p_help = pred[2]
            p_harm = pred[3]

            effort = 1 - p_wait
            competence = pred[0] / p_try if p_try > 0.01 else 0.5
            warmth = p_help - p_harm

            return {"effort": effort, "competence": competence, "warmth": warmth}
    return (SlidingWindowHMM,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## State Machine Visualization (Mode-Centric)

    Nodes are **modes**, edges show transitions after specific outcomes.

    Treating "try" as a composite outcome (success/fail weighted by competence).
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    outcome_selector = mo.ui.dropdown(
        options=["success", "fail", "help", "harm", "wait", "try"],
        value="help",
        label="Outcome to visualize",
        full_width=False
    )

    threshold_slider = mo.ui.slider(
        start=0.0,
        stop=0.5,
        step=0.025,
        value=0.1,
        label="Minimum probability threshold",
        full_width=False
    )

    mo.hstack([outcome_selector, threshold_slider])
    return outcome_selector, threshold_slider


@app.cell(hide_code=True)
def _(outcome_selector, threshold_slider, visualize_state_machine):
    visualize_state_machine(
        outcome_filter=outcome_selector.value,
        threshold=threshold_slider.value
    )
    return


@app.cell(hide_code=True)
def _(Agent, defaultdict, mode_colors, nx, outcome_colors, plt):
    def visualize_state_machine(outcome_filter, threshold=0.1):
        """
        Visualize the Agent's state transitions as a directed graph.

        Edges point FROM current_mode TO next_mode after outcome occurs.
        Example: prosocial --[success]--> prosocial (self-loop with 0.85 probability)

        Args:
            outcome_filter: Which outcome to show ("success", "fail", "help", "harm", "wait", "try")
            threshold: Minimum probability to display (0-1)
        """
        agent = Agent()

        G = nx.DiGraph()

        for mode in ["prosocial", "antisocial", "withdrawn"]:
            G.add_node(mode)

        edge_data = defaultdict(list)

        if outcome_filter == "try":
            # "try" combines success and fail weighted by competence
            for from_mode in ["prosocial", "antisocial", "withdrawn"]:
                comp = agent.competence[from_mode]
                success_trans = agent.transitions[(from_mode, "success")]
                fail_trans = agent.transitions[(from_mode, "fail")]

                for to_mode in ["prosocial", "antisocial", "withdrawn"]:
                    # P(to_mode | try) = P(success) * P(to_mode | success) + P(fail) * P(to_mode | fail)
                    prob = comp * success_trans[to_mode] + (1 - comp) * fail_trans[to_mode]
                    if prob < threshold:
                        continue

                    edge = (from_mode, to_mode)
                    edge_data[edge].append(("try", prob, outcome_colors.get("try", "#7f8c8d"), prob * 5))
        else:
            for (from_mode, outcome), trans_probs in agent.transitions.items():
                if outcome != outcome_filter:
                    continue

                for to_mode, prob in trans_probs.items():
                    if prob < threshold:
                        continue

                    edge = (from_mode, to_mode)
                    edge_data[edge].append((outcome, prob, outcome_colors.get(outcome, "#7f8c8d"), prob * 5))

        for edge in edge_data:
            G.add_edge(*edge)

        fig, ax = plt.subplots(figsize=(14, 10))
        pos = nx.circular_layout(G)

        nx.draw_networkx_nodes(G, pos, node_color=[mode_colors[node] for node in G.nodes()], 
                               node_size=4000, alpha=0.9, ax=ax)

        nx.draw_networkx_labels(G, pos, font_size=12, font_weight='bold', 
                                font_color='white', ax=ax)

        edges_by_style = {}
        connectionstyle = 'arc3,rad=0.25'

        for edge, outcomes_list in edge_data.items():
            avg_width = sum(d[3] for d in outcomes_list) / len(outcomes_list)
            max_idx = max(range(len(outcomes_list)), key=lambda i: outcomes_list[i][1])
            edge_color = outcomes_list[max_idx][2]

            if connectionstyle not in edges_by_style:
                edges_by_style[connectionstyle] = []
            edges_by_style[connectionstyle].append((edge, avg_width, edge_color, outcomes_list))

        for connectionstyle, edge_list in edges_by_style.items():
            nx.draw_networkx_edges(
                G, pos, 
                edgelist=[e[0] for e in edge_list],
                width=[e[1] for e in edge_list], 
                alpha=0.7,
                edge_color=[e[2] for e in edge_list], 
                connectionstyle=connectionstyle,
                arrows=True, 
                arrowsize=30, 
                arrowstyle='-|>',
                node_size=4000,
                ax=ax,
                min_source_margin=20,
                min_target_margin=20
            )

        edge_labels = {}
        for edge, outcomes_list in edge_data.items():
            non_zero = [(outcome, prob) for outcome, prob, _, _ in outcomes_list if prob > 0]
            if non_zero:
                label = "\n".join([f"{prob:.2f}" for outcome, prob in non_zero])
                edge_labels[edge] = label

        label_bbox = dict(boxstyle='round,pad=0.3', facecolor='white', alpha=0.9, edgecolor='gray', linewidth=0.5)
        nx.draw_networkx_edge_labels(
            G, pos, 
            edge_labels=edge_labels,
            font_size=8,
            bbox=label_bbox,
            connectionstyle=connectionstyle,
            ax=ax
        )

        title = f"State Machine: '{outcome_filter}' Outcome (threshold={threshold:.3f})"
        ax.set_title(title, fontsize=14, fontweight='bold', pad=20)

        ax.axis('off')
        ax.set_xlim(-1., 1.4)
        ax.set_ylim(-1.4, 1.4)
        plt.tight_layout()

        return fig
    return (visualize_state_machine,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Action Transition Graph (Action-Centric)

    Nodes are **actions/outcomes**, edges show action-to-action transitions.
    The mode filter controls which behavioral context to visualize.

    Treating "try" as an intermediate action (leads to success/fail based on competence).
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mode_selector = mo.ui.dropdown(
        options=["prosocial", "antisocial", "withdrawn"],
        value="prosocial",
        label="Mode",
        full_width=False
    )
    mode_selector
    return (mode_selector,)


@app.cell
def _(mode_selector, visualize_action_graph):
    visualize_action_graph(mode_filter=mode_selector.value)
    return


@app.cell(hide_code=True)
def _(Agent, mode_colors, nx, outcome_colors, plt):
    def visualize_action_graph(mode_filter="prosocial"):
        """
        Visualize actions as nodes for a single mode (assuming mode doesn't change).

        Edge weights are P(next_action | mode) - the probability of each action in the mode.
        All outgoing edges from any node have the same weights since mode is fixed.

        Args:
            mode_filter: Which mode to show ("prosocial", "antisocial", "withdrawn")
        """
        agent = Agent()

        mode_actions = agent.modes[mode_filter]
        p_success = agent.competence[mode_filter]

        actions = ["help", "harm", "try", "wait"]
        outcomes = ["success", "fail"]

        # Build edge data: (from, to) -> prob, filtering out zero probabilities
        edge_data = {}

        # try -> success/fail based on competence
        edge_data[("try", "success")] = p_success
        edge_data[("try", "fail")] = 1 - p_success

        # All other transitions: from any action/outcome to the next action
        for from_node in actions + outcomes:
            if from_node == "try":
                continue
            for next_action, prob in mode_actions.items():
                if prob > 0:
                    edge_data[(from_node, next_action)] = prob

        # Build graph with only non-zero edges
        G = nx.DiGraph()
        for action in actions + outcomes:
            G.add_node(action)
        for edge in edge_data:
            G.add_edge(*edge)

        fig, ax = plt.subplots(figsize=(14, 10))
        pos = nx.spring_layout(G, k=2, iterations=50, seed=42)

        nx.draw_networkx_nodes(G, pos, node_color=[outcome_colors[node] for node in G.nodes()],
                               node_size=3000, alpha=0.9, ax=ax)

        nx.draw_networkx_labels(G, pos, font_size=11, font_weight='bold',
                                font_color='white', ax=ax)

        # Draw all edges at once
        edge_color = mode_colors[mode_filter]
        connectionstyle = 'arc3,rad=0.2'

        nx.draw_networkx_edges(
            G, pos,
            edgelist=list(edge_data.keys()),
            width=[5 * edge_data[e] for e in edge_data],
            alpha=0.7,
            edge_color=edge_color,
            connectionstyle=connectionstyle,
            arrows=True,
            arrowsize=20,
            arrowstyle='-|>',
            node_size=3000,
            ax=ax
        )

        # Edge labels for non-zero edges only
        edge_labels = {e: f"{p:.2f}" for e, p in edge_data.items()}

        label_bbox = dict(boxstyle='round,pad=0.2', facecolor='white', alpha=0.8, edgecolor='gray', linewidth=0.5)
        nx.draw_networkx_edge_labels(
            G, pos,
            edge_labels=edge_labels,
            font_size=9,
            bbox=label_bbox,
            connectionstyle=connectionstyle,
            ax=ax
        )

        title = f"Action Transition Graph: {mode_filter.capitalize()} Mode"
        ax.set_title(title, fontsize=14, fontweight='bold', pad=20)

        ax.axis('off')
        ax.set_xlim(-1.4, 1.4)
        ax.set_ylim(-1.4, 1.4)
        plt.tight_layout()

        return fig
    return (visualize_action_graph,)


@app.cell(hide_code=True)
def _(plot_comparison, results):
    # Plot
    plot_comparison(results)
    return


@app.cell(hide_code=True)
def _(Counter, results):
    # Count tokens
    counts = Counter(results['tokens'])
    print("\nSequence statistics:")
    for t, c in counts.most_common():
        print(f"  {t}: {c}")

    # Count modes
    mode_counts = Counter(results['true_modes'])
    print("\nTrue mode frequencies:")
    for m, c in mode_counts.most_common():
        print(f"  {m}: {c} ({c/len(results['true_modes'])*100:.1f}%)")
    return


@app.cell(hide_code=True)
def _(np, run_comparison):
    np.random.seed(4)
    results = run_comparison(seq_length=150, window_size=25)
    return (results,)


@app.cell
def _(Agent, Counter, SlidingWindowHMM, TOKEN_COLORS, mpatches, np, plt):
    def run_comparison(seq_length=150, window_size=20):    
        # Generate sequence
        agent = Agent()
        agent.reset()

        tokens = []
        true_modes = []
        for _ in range(seq_length):
            token, mode = agent.generate_token()
            tokens.append(token)
            true_modes.append(mode)

        # Initialize models
        bayesian = OnlineBayesian()
        hmm_model = SlidingWindowHMM(window_size=window_size, n_states=3)

        # Track estimates over time
        bayes_effort, bayes_comp, bayes_warmth = [], [], []
        hmm_effort, hmm_comp, hmm_warmth = [], [], []

        for token in tokens:
            bayesian.update(token)
            hmm_model.update(token)

            b_traits = bayesian.get_traits()
            h_traits = hmm_model.get_traits()

            bayes_effort.append(b_traits['effort'])
            bayes_comp.append(b_traits['competence'])
            bayes_warmth.append(b_traits['warmth'])

            hmm_effort.append(h_traits['effort'])
            hmm_comp.append(h_traits['competence'])
            hmm_warmth.append(h_traits['warmth'])

        # Compute "true" local traits (from sliding window empirical counts)
        true_effort, true_comp, true_warmth = [], [], []
        for i in range(len(tokens)):
            window = tokens[max(0, i-window_size+1):i+1]
            counts = Counter(window)
            n = len(window)

            n_wait = counts.get("wait", 0)
            n_try_s = counts.get("try_success", 0)
            n_try_f = counts.get("try_fail", 0)
            n_help = counts.get("help", 0)
            n_harm = counts.get("harm", 0)

            true_effort.append(1 - n_wait/n)
            true_comp.append(n_try_s / (n_try_s + n_try_f) if (n_try_s + n_try_f) > 0 else 0.5)
            true_warmth.append((n_help - n_harm) / n)

        return {
            'tokens': tokens,
            'true_modes': true_modes,
            'bayes': {'effort': bayes_effort, 'competence': bayes_comp, 'warmth': bayes_warmth},
            'hmm': {'effort': hmm_effort, 'competence': hmm_comp, 'warmth': hmm_warmth},
            'local_truth': {'effort': true_effort, 'competence': true_comp, 'warmth': true_warmth},
            'window_size': window_size
        }


    def plot_comparison(results):
        tokens = results['tokens']
        true_modes = results['true_modes']
        bayes = results['bayes']
        hmm = results['hmm']
        local = results['local_truth']
        window_size = results['window_size']

        fig, axes = plt.subplots(5, 1, figsize=(16, 13), sharex=True,
                                 gridspec_kw={'height_ratios': [0.6, 0.6, 1, 1, 1]})
        steps = np.arange(1, len(tokens) + 1)

        # Panel 1: Action sequence
        ax = axes[0]
        for i, token in enumerate(tokens):
            color = TOKEN_COLORS.get(token, "#bdc3c7")
            ax.barh(0, 1, left=i, color=color, edgecolor='none')
        ax.set_xlim(0, len(tokens))
        ax.set_ylim(-0.5, 0.5)
        ax.set_yticks([])
        ax.set_title("Observed Action Sequence", fontsize=11, fontweight='bold')
        patches = [mpatches.Patch(color=c, label=t) for t, c in TOKEN_COLORS.items()]
        ax.legend(handles=patches, loc='upper right', ncol=5, fontsize=8)
        # Add small markers on the action sequence for each "try" event
        try_indices = [i for i, token in enumerate(tokens) if token.startswith("try_")]
        ax.scatter(try_indices, [0.] * len(try_indices), 
                  marker='|', s=50, color='#e67e22', alpha=1., zorder=10, clip_on=False)

        # Panel 2: True mode
        ax = axes[1]
        mode_colors = {"prosocial": "#2ecc71", "antisocial": "#e74c3c", "withdrawn": "#95a5a6"}
        for i, mode in enumerate(true_modes):
            ax.barh(0, 1, left=i, color=mode_colors[mode], edgecolor='none')
        ax.set_xlim(0, len(tokens))
        ax.set_ylim(-0.5, 0.5)
        ax.set_yticks([])
        ax.set_title("Agent's TRUE Internal Mode (Hidden from Observer)", fontsize=11, fontweight='bold')
        patches = [mpatches.Patch(color=c, label=m) for m, c in mode_colors.items()]
        ax.legend(handles=patches, loc='upper right', ncol=3, fontsize=9)

        # Apply light smoothing to HMM for readability
        def smooth(x, w=3):
            return np.convolve(x, np.ones(w)/w, mode='same')

        hmm_effort_smooth = smooth(hmm['effort'], 5)
        hmm_comp_smooth = smooth(hmm['competence'], 5)
        hmm_warmth_smooth = smooth(hmm['warmth'], 5)

        # Panel 3: Effort
        ax = axes[2]
        ax.fill_between(steps, local['effort'], alpha=0.15, color='gray')
        ax.plot(steps, local['effort'], color='gray', linewidth=1, alpha=0.5, label=f'Local empirical (window={window_size})')
        ax.plot(steps, bayes['effort'], color='#c0392b', linewidth=2.5, label='Bayesian (cumulative)')
        ax.plot(steps, hmm_effort_smooth, color='#2980b9', linewidth=2.5, label=f'HMM (window={window_size})')
        ax.set_ylim(0.2, 1.05)
        ax.set_ylabel("Effort", fontsize=11)
        ax.set_title("EFFORT: 1 - P(wait)", 
                     fontsize=11, fontweight='bold')
        ax.legend(loc='lower right', fontsize=9)
        ax.grid(True, alpha=0.3)

        # Panel 4: Competence
        ax = axes[3]
        ax.fill_between(steps, local['competence'], alpha=0.15, color='gray')
        ax.plot(steps, local['competence'], color='gray', linewidth=1, alpha=0.5, label='Local empirical')
        ax.plot(steps, bayes['competence'], color='#c0392b', linewidth=2.5, label='Bayesian')
        ax.plot(steps, hmm_comp_smooth, color='#2980b9', linewidth=2.5, label='HMM')
        ax.set_ylim(0, 1.05)
        ax.set_ylabel("Competence", fontsize=11)
        ax.set_title("COMPETENCE: P(success | try)", 
                     fontsize=11, fontweight='bold')
        ax.legend(loc='lower right', fontsize=9)
        ax.axhline(0.5, color='gray', linestyle='--', alpha=0.3)
        ax.grid(True, alpha=0.3)

        # Panel 5: Warmth - most diagnostic!
        ax = axes[4]
        ax.fill_between(steps, local['warmth'], alpha=0.15, color='gray')
        ax.plot(steps, local['warmth'], color='gray', linewidth=1, alpha=0.5, label='Local empirical')
        ax.plot(steps, bayes['warmth'], color='#c0392b', linewidth=2.5, label='Bayesian')
        ax.plot(steps, hmm_warmth_smooth, color='#2980b9', linewidth=2.5, label='HMM')
        ax.set_ylim(-0.7, 0.7)
        ax.set_ylabel("Warmth", fontsize=11)
        ax.set_xlabel("Time step", fontsize=11)
        ax.set_title("WARMTH: P(help) - P(harm) ", 
                     fontsize=11, fontweight='bold')
        ax.legend(loc='lower right', fontsize=9)
        ax.axhline(0, color='gray', linestyle='--', alpha=0.5)
        ax.grid(True, alpha=0.3)

        # Add shaded regions for mode phases
        for ax in axes[2:]:
            for i, mode in enumerate(true_modes):
                if mode == "antisocial":
                    ax.axvspan(i, i+1, alpha=0.08, color='red', linewidth=0)
                elif mode == "withdrawn":
                    ax.axvspan(i, i+1, alpha=0.08, color='gray', linewidth=0)

        fig.suptitle("Sliding Window HMM vs Online Bayesian Model: Tracking Behavioral Phases", 
                     fontsize=13, fontweight='bold', y=0.99)
        plt.tight_layout()

        return fig
    return plot_comparison, run_comparison


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "marimo>=0.19.0",
#     "pyzmq",
# ]
# ///

# NOTE: Use the command palette (cmd+shift+p) and search "marimo start in editor" to generate the live preview.

import marimo

__generated_with = "0.19.11"
app = marimo.App(width="full")

with app.setup:
    import numpy as np
    from mcp_hmm.hmm_xiang2023 import (
        generate_windows,
        generate_outcome,
        WindowableHMM,
        DebuggingMonitor,
        plot_results,
        plot_hmm_matrices,
        visualize_agent_contributions_sampdist,
        AgentPair,
        TOKENS,
        TOKEN_COLORS,
        make_null_actions_generator,
        generate_windows_with_actions_generator
    )
    from mcp_hmm.hmm_xiang2023_bwd_meta import (
        show_accuracy,
        order_hiddens_by_emission_probs,
        visualize_markov_chain,
        make_heatmap_trajectory_inspector,
        make_markov_chain_trajectory_inspector,
        tidy_scores,
        infer_competence_from_lift_prob,
        visualize_competence_results,
        get_window_types
    )

    import networkx as nx
    import polars as pl
    from polars import lit, col
    import marimo as mo

    from sklearn.model_selection import KFold, LeaveOneGroupOut

    import seaborn as sns
    from matplotlib import pyplot as plt
    from scipy.optimize import minimize


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Cross-validation
    """)
    return


@app.cell(hide_code=True)
def _():
    # discriminability = mo.ui.slider(start=0, stop=1, value=.5, step=.1, label="Discriminability")
    discriminability_noise = mo.ui.slider(start=0, stop=1, value=.2, step=.1, label="Discriminability Noise")
    A_competence = mo.ui.slider(start=0, stop=10, value=4, step=.5, label="A Competence")
    B_competence = mo.ui.slider(start=0, stop=10, value=6, step=.5, label="B Competence")

    mo.vstack([ discriminability_noise, A_competence, B_competence ])
    return A_competence, B_competence, discriminability_noise


@app.cell(hide_code=True)
def _(A_competence, B_competence, discriminability_noise):
    np.random.seed(4)
    pair = AgentPair(discriminability_noise=discriminability_noise.value, A_competence=A_competence.value, B_competence=B_competence.value)

    visualize_agent_contributions_sampdist(pair)
    return (pair,)


@app.cell(hide_code=True)
def _(accuracy_results):
    show_accuracy(accuracy_results)
    return


@app.cell(hide_code=True)
def _(pair):
    n_windows = 100

    np.random.seed(4)
    fwd_windows = np.array(generate_windows(pair, num_windows=n_windows))
    _fwd_windows_types = get_window_types(fwd_windows)

    test_scores = []
    train_scores = []
    # test_scores_b = []
    # train_scores_b = []
    for i, (train, test) in enumerate(LeaveOneGroupOut().split(fwd_windows, groups=_fwd_windows_types)):
        np.random.seed(i)

        train_windows = fwd_windows[train]
        test_windows = fwd_windows[test]

        fwd_model = WindowableHMM(n_states=10, max_iter=120)
        for win in train_windows:
            fwd_model.add_to_history(win)
        fwd_model.fit()

        print(f"{fwd_model.model.monitor_.iter} iterations out of {fwd_model.model.monitor_.n_iter} max")

        train_scores.append(fwd_model.accuracy(train_windows).reshape(-1))
        test_scores.append(fwd_model.accuracy(test_windows).reshape(-1))

        # train_scores_b.append(fwd_model.accuracy_about_stronger_agent(train_windows).reshape(-1))
        # test_scores_b.append(fwd_model.accuracy_about_stronger_agent(test_windows).reshape(-1))
    return fwd_windows, n_windows, test_scores, train_scores


@app.cell(hide_code=True)
def _(test_scores, train_scores):
    accuracy_results = pl.concat([
        tidy_scores(train_scores).with_columns(label=lit('train'), target=lit('outcome')),
        tidy_scores(test_scores).with_columns(label=lit('test'), target=lit('outcome')),
        # tidy_scores(train_scores_b).with_columns(label=lit('train'), target=lit('stronger agent')),
        # tidy_scores(test_scores_b).with_columns(label=lit('test'), target=lit('stronger agent'))
    ])
    return (accuracy_results,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Inspecting trajectory of hidden weights during fit
    """)
    return


@app.cell(hide_code=True)
def _(fwd_windows):
    np.random.seed(2)
    _n_states = 10
    # TI = trajectory inspector
    TI_fwd_model = WindowableHMM(n_states=_n_states, max_iter=120, save_props_while_fitting=['startprob_', 'transmat_', 'emissionprob_'], custom_model_args={
        # randomize the transmat priors to try to help more varied solutions arise (randomizing the startprob_prior doesn't seem to have an impacts, and can only initialize all the emission probs to the same value)
        'transmat_prior': np.random.rand(_n_states, _n_states)
    })
    for _win in fwd_windows:
        TI_fwd_model.add_to_history(_win)
    TI_fwd_model.fit()
    return (TI_fwd_model,)


@app.cell
def _(TI_fwd_model):
    hmTI, hmTI_frame_slider = make_heatmap_trajectory_inspector(TI_fwd_model)
    return hmTI, hmTI_frame_slider


@app.cell(hide_code=True)
def _(hmTI, hmTI_frame_slider):
    hmTI(hmTI_frame_slider)
    return


@app.cell(hide_code=True)
def _(TI_fwd_model):
    mcTI, a, mcTI_threshold_slider = make_markov_chain_trajectory_inspector(TI_fwd_model)
    return a, mcTI, mcTI_threshold_slider


@app.cell(hide_code=True)
def _(a, mcTI, mcTI_threshold_slider):
    mcTI(a, mcTI_threshold_slider)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Inferring competence from learned transition probabilities

    Our model has that:
    - P(lift | X_try, Y_try) ~ (Competence(X) + Competence(Y)) >= 5
    - P(lift | X_try, Y_wait) ~ Competence(X) >= 5
    - P(lift | X_wait, Y_wait) ~ 0
    - P(fail | ...) ~ 1 - P(lift | ...)

    with X and Y being A and B in any order,  with the conditioning events in no particular order, and ... being any of the three conditions. Importantly, `discriminability_noise` doesn't change these relationships, it just introduces noise.

    So then let Competence(X) be inferred by Bayesian inference on the probabilities learned by the HMM (where X can be A or B). Since the HMM is forwards-learning (i.e., sequences are [action1, action2, outcome]), we can directly estimate P(lift | actions) and P(fail | actions) by conditioning on the observed actions and reading off the predicted emission for the outcome token. No Bayes' rule inversion is needed (unlike the backwards-learning model).
    """)
    return


@app.function
# Step 1: Extract P(lift | actions) directly from the forward HMM
# The forward model sees [action1, action2, outcome].
# We condition on the two actions and read off P(outcome) directly.

def infer_lift_prob(fwd_model, pair):
    _model = fwd_model.model
    _tok2idx = fwd_model.token_to_idx

    def _p_outcome_given_actions(a1, a2):
        """Compute P(outcome=lift) and P(outcome=fail) given observing [a1, a2]."""
        _obs = fwd_model._tokens_to_obs([a1, a2])
        _, _posteriors = _model.score_samples(_obs, [len(_obs)])
        _state_dist_after_actions = _posteriors[-1]

        # One more transition + emission to get P(next token = outcome)
        _next_state_dist = _state_dist_after_actions @ _model.transmat_
        _p = {}
        for _outcome in ["lift", "fail"]:
            _oidx = _tok2idx[_outcome]
            _p[_outcome] = (_next_state_dist * _model.emissionprob_[:, _oidx]).sum()
        return _p

    # Conditions: unordered action pairs
    _conditions = [
        ("A_try", "B_try"),
        ("A_try", "B_wait"),
        ("B_try", "A_wait"),
        ("A_wait", "B_wait"),
    ]

    _hmm_p_lift = {}
    for _a1, _a2 in _conditions:
        _label = f"{_a1}, {_a2}"

        # ordering 1: a1 then a2
        _p1 = _p_outcome_given_actions(_a1, _a2)

        if _a1 != _a2:
            # ordering 2: a2 then a1 (actions are unordered in the domain)
            _p2 = _p_outcome_given_actions(_a2, _a1)
            # Average P(lift) across both orderings, weighted by their total mass
            _total_lift = _p1["lift"] + _p2["lift"]
            _total_all = _total_lift + _p1["fail"] + _p2["fail"]
        else:
            _total_lift = _p1["lift"]
            _total_all = _total_lift + _p1["fail"]

        _hmm_p_lift[_label] = _total_lift / _total_all if _total_all > 0 else float("nan")

    # Marginal P(outcome) for display (average over all conditions)
    _p_outcome = {"lift": 0.0, "fail": 0.0}
    for _label in _hmm_p_lift:
        _p_outcome["lift"] += _hmm_p_lift[_label]
        _p_outcome["fail"] += 1.0 - _hmm_p_lift[_label]
    _total = _p_outcome["lift"] + _p_outcome["fail"]
    _p_outcome = {k: v / _total for k, v in _p_outcome.items()}

    # Ground truth via Monte Carlo
    _n_mc = 10000
    _gt_p_lift = {}
    for _a1, _a2 in _conditions:
        _label = f"{_a1}, {_a2}"
        _lifts = 0
        for _ in range(_n_mc):
            _lifts += 1 if generate_outcome(pair, _a1, _a2) == "lift" else 0
        _gt_p_lift[_label] = _lifts / _n_mc

    lift_prob = {
        "conditions": [f"{a1}, {a2}" for a1, a2 in _conditions],
        "hmm_p_lift": _hmm_p_lift,
        "gt_p_lift": _gt_p_lift,
        "p_outcome": _p_outcome,
        "noise": 0.5,
        # ^ don't just feed in discriminability_noise.value -- it should be inferred by the participant model from the data, OR a free parameter set to optimize fit to real participant data. it's cheating if the model of the participant knows generation parameters! should only have access to observations
    }
    return lift_prob


@app.cell
def _(TI_fwd_model, pair):
    lift_prob = infer_lift_prob(TI_fwd_model, pair)
    return (lift_prob,)


@app.cell
def _(lift_prob, pair):
    # Step 2: Comparison table + competence inversion
    _lp = lift_prob

    _comparison_df = pl.DataFrame({
        "Condition": _lp["conditions"],
        "HMM P(lift)": [_lp["hmm_p_lift"][c] for c in _lp["conditions"]],
        "Ground Truth P(lift)": [_lp["gt_p_lift"][c] for c in _lp["conditions"]],
    }).with_columns(
        Difference=(col("HMM P(lift)") - col("Ground Truth P(lift)"))
    )

    # Invert solo-try conditions to estimate competence
    # Effort ~ Beta(alpha, 1), so p_lift = P(C*E >= 5) = 1 - (5/C)^alpha when C > 5
    # Inverting: C = 5 / (1 - p_lift)^(1/alpha)
    _noise = _lp["noise"]
    _alpha = 1.0 / max(_noise**2, 0.00001)

    def _invert_competence(p_lift): 
        if p_lift == 0:
            return None  # can only say C <= 5
        if p_lift == 1:
            return float("inf")  # always succeeds solo, C unbounded above
        return 5.0 / (1.0 - p_lift) ** (1.0 / _alpha)

    _agents = ["A", "B"]
    _true_comp = [pair.A_competence, pair.B_competence]
    _solo_conditions = ["A_try, B_wait", "B_try, A_wait"]
    _hmm_inferred = [_invert_competence(_lp["hmm_p_lift"][c]) for c in _solo_conditions]
    _gt_inferred = [_invert_competence(_lp["gt_p_lift"][c]) for c in _solo_conditions]

    _competence_rows = []
    for _i, _agent in enumerate(_agents):
        _competence_rows.append({
            "Agent": _agent,
            "True Competence": _true_comp[_i],
            "GT-Inferred": _gt_inferred[_i] if _gt_inferred[_i] is not None else float("nan"),
            "HMM-Inferred": _hmm_inferred[_i] if _hmm_inferred[_i] is not None else float("nan"),
            "Solo P(lift) [HMM]": _lp["hmm_p_lift"][_solo_conditions[_i]],
            "Solo P(lift) [GT]": _lp["gt_p_lift"][_solo_conditions[_i]],
        })
    competence_results = pl.DataFrame(_competence_rows)

    mo.vstack([
        mo.md(f"""
    ### P(lift | actions): HMM vs Ground Truth

    HMM marginal: P(lift) = {_lp['p_outcome']['lift']:.3f}, P(fail) = {_lp['p_outcome']['fail']:.3f}

    Effort distribution: Beta(alpha={_alpha:.1f}, 1), noise={_noise}
        """),
        _comparison_df,
        mo.md("### Competence inversion from solo-try conditions"),
        mo.md(f"Formula: C = 5 / (1 - P(lift))^(1/{_alpha:.1f}). NaN means P(lift) = 0, so C <= 5."),
        competence_results,
    ])
    return


@app.cell
def _(lift_prob, pair):
    # Show predicted vs observed lift probs for the joint fit
    np.random.seed(0)
    joint_competence_results, original_predicted_p, optimized_predicted_p = infer_competence_from_lift_prob(lift_prob, pair)

    mo.vstack([
        mo.md(f"""
    ### Joint competence inference (using solo + both-try conditions)

    Jointly optimizes (C_A, C_B) to fit three HMM-derived lift probabilities:
    - P(lift | A_try, B_wait) = {original_predicted_p['A_solo']:.3f} (predicted: {optimized_predicted_p['A_solo']:.3f})
    - P(lift | B_try, A_wait) = {original_predicted_p['B_solo']:.3f} (predicted: {optimized_predicted_p['B_solo']:.3f})
    - P(lift | A_try, B_try) = {original_predicted_p['joint']:.3f} (predicted: {optimized_predicted_p['joint']:.3f})
        """),
        joint_competence_results,
    ])
    return (joint_competence_results,)


@app.cell
def _(joint_competence_results, pair):
    visualize_competence_results(joint_competence_results, pair)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Cross-prediction from null to intact data

    Null data so defined that this tests that the model needs to learn the temporal contingencies of the actions.
    """)
    return


@app.cell
def _(fwd_windows, n_windows, pair):
    null_actions_gen, _, _ = make_null_actions_generator()

    np.random.seed(4)
    null_fwd_windows = np.array(generate_windows_with_actions_generator(pair, n_windows, null_actions_gen))
    _null_fwd_windows_types = get_window_types(null_fwd_windows)

    _test_scores = []
    _train_scores = []
    # _test_scores_b = []
    # _train_scores_b = []
    for _i, (_train, _test) in enumerate(LeaveOneGroupOut().split(null_fwd_windows, groups=_null_fwd_windows_types)):
        np.random.seed(_i)

        _train_windows = null_fwd_windows[_train]

        _rng = np.random.default_rng()
        _test_scores_iteration_means = []
        for _j in range(10):
            print(f"Fold {_i}, Iteration {_j}")
            _test_windows = _rng.choice(fwd_windows, size=len(_test), replace=False)
            # ^ could use ALL intact data, but for the sake of matching the precision of the scores to the regular version, sample the same number as in the selected test set.

            _fwd_model = WindowableHMM(n_states=10, max_iter=120)
            for _win in _train_windows:
                _fwd_model.add_to_history(_win)
            _fwd_model.fit()

            print(f"{_fwd_model.model.monitor_.iter} iterations out of {_fwd_model.model.monitor_.n_iter} max")

            _test_scores_iteration_means.append(_fwd_model.accuracy(_test_windows).reshape(-1).mean())

        _train_scores.append(_fwd_model.accuracy(_train_windows).reshape(-1))
        _test_scores.append(_test_scores_iteration_means)

    null_cp_accuracy_results = pl.concat([
        tidy_scores(_train_scores).with_columns(label=lit('train'), target=lit('outcome')),
        tidy_scores(_test_scores).with_columns(label=lit('test'), target=lit('outcome')),
        # tidy_scores(_train_scores_b).with_columns(label=lit('train'), target=lit('stronger agent')),
        # tidy_scores(_test_scores_b).with_columns(label=lit('test'), target=lit('stronger agent'))
    ])
    return (null_cp_accuracy_results,)


@app.cell
def _(null_cp_accuracy_results):
    show_accuracy(null_cp_accuracy_results)
    return


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()



# This is a Python file that's designed to be run as a script
# You run it using the following command:
# `pixi run python code/demo.py`
#
# Or using the pre-configured pixi task:
# `pixi run demo`
import sys


def calc_age(age: int) -> tuple[int, int, int, int]:
    """Calculate the age of person in number of months, weeks, days, and minutes.

    Args:
        age (int): Age in years

    Returns:
        tuple[int, int, int, int]: Age in months, weeks, days, and minutes
    """
    return age * 12, age * 52, age * 365, age * 365 * 24 * 60


def main(name: str, age: int | None = None):
    """The main function of the demo.py script.

    Args:
        name (str): Name of the person
        age (int | None, optional): Age of the person. Defaults to None.
    """
    print(f"Hello from the demo.py script {name}!")
    if age:
        months, weeks, days, minutes = calc_age(age)
        print(
            f"Your age in:\n months: {months}\n weeks: {weeks}\n days: {days}\n minutes: {minutes}\n"
        )


if __name__ == "__main__":
    # sys.argv[] is a list of any arguments passed
    # sys.argv[0] is the name of the script
    name = sys.argv[1] if len(sys.argv) > 1 else ""
    age = int(sys.argv[2]) if len(sys.argv) > 2 else None
    main(name, age)
